from urllib.parse import parse_qs, urlparse

import requests

base_url = "https://auth.bullhornstaffing.com/oauth"


def get_auth_code(username, password, client_id):
    """
    Retrieve the authorization code by initiating the OAuth flow.
    """
    data = {
        "client_id": client_id,
        "response_type": "code",
        "username": username,
        "password": password,
        "action": "Login",
    }
    authorize_url = base_url + "/authorize"
    response = requests.post(authorize_url, data=data, allow_redirects=True)
    if response.ok:
        redirect_url = response.url
        parsed_url = urlparse(redirect_url)
        auth_code = parse_qs(parsed_url.query)["code"][0]
        return auth_code
    raise Exception(
        f"Authorization failed with status code {response.status_code}: {response.text}"
    )


def make_token_request(data):
    """
    Make a request to obtain the OAuth access token.
    """
    token_url = base_url + "/token"
    response = requests.post(token_url, data=data)
    if response.ok:
        return response.json()

    raise Exception(
        f"Token request failed with status code {response.status_code}: {response.text}"
    )


def login_to_bullhorn(access_token):
    """
    Log in to Bullhorn using the obtained access token.
    """
    login_url = "https://rest.bullhornstaffing.com/rest-services/login"
    params = {"version": "2.0", "access_token": access_token["access_token"]}
    response = requests.post(url=login_url, params=params)

    if response.ok:
        auth_response = response.json()
        auth_response["refresh_token"] = access_token["refresh_token"]
        return auth_response

    raise Exception(
        f"Login to Bullhorn failed with status code {response.status_code}:"
        f" {response.text}"
    )


def get_or_refresh_token(
    grant_type, client_id, client_secret, ttl=None, code=None, refresh_token=None
):
    """
    Gets or refreshes an OAuth access token based on the grant type.
    """
    data = {
        "grant_type": grant_type,
        "client_id": client_id,
        "client_secret": client_secret,
    }
    if grant_type == "authorization_code":
        data["code"] = code
    elif grant_type == "refresh_token":
        data["refresh_token"] = refresh_token

    # Add TTL if specified
    if ttl:
        data["ttl"] = ttl

    token_response = make_token_request(data)
    # Login to Bullhorn and return the response
    return login_to_bullhorn(token_response)


def auth(
    username, password, client_id, client_secret, refresh_token=None, auth_code=None
):
    """
    Obtain the access token for authentication.
    """
    if refresh_token:
        access_token = get_or_refresh_token(
            "refresh_token",
            client_id,
            client_secret,
            ttl=604800,
            refresh_token=refresh_token,
        )
    elif auth_code:
        access_token = get_or_refresh_token(
            "authorization_code", client_id, client_secret, ttl=604800, code=auth_code
        )  # 7 days in seconds
    else:
        auth_code = get_auth_code(username, password, client_id)
        access_token = get_or_refresh_token(
            "authorization_code", client_id, client_secret, ttl=604800, code=auth_code
        )
    return access_token



import typing as t
from enum import Enum, IntEnum
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec import json as msgspec_json
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.francetravail.schemas import (
    ExperienceRequirement,
    FranceTravailJobOffer,
    OfferOriginTag,
    validate_date,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    Endpoint,
    Endpoints,
    ReadOperation,
    merge,
)

API_BASE_URL = "https://api.francetravail.io/partenaire/offresdemploi"
JOBS_SEARCH_URL = ("{}/v2/offres/search").format(API_BASE_URL)

REFERENCES_URL = ("{}/v2/referentiel/").format(API_BASE_URL)

SEARCH_JOBS_ENDPOINT = Endpoint(
    name="Get all jobs",
    description=(
        "Endpoint to search offers based on selection criteria. The list of returned"
        " offers is paginated.The range of results is limited to 150.The request method"
        " is `GET`"
    ),
    url=JOBS_SEARCH_URL,
)
TOKEN_GENERATOR_URL = (
    "https://entreprise.francetravail.fr/connexion/oauth2/access_token"
)

GRANT_TYPE = "client_credentials"
TOKEN_SCOPE = "api_offresdemploiv2 o2dsoffre"

MOTS_CLES_DESCRIPTION = """Search by keyword

Each keyword (or expression) is at least 2 characters long and must
be separated by a comma.
The search on several keywords is processed via the logical operator "AND".
The keyword search can be used to search on :

- The title of the offer (title field in the search return)
- The ROME code (romeCode field in the search return)
- The ROME label (field romeLibelle in return for the search)
- The competences label (field competences.libelle in return of the search)
- The wording of the training fields (field formations.domaineLibelle in
return of the research)
- The wording of the permits (field permits.label in return of the search)
- The language label (field languages.label in return of the search)
- The offer description if found in the offer title and/or the ROME label
(description field in the search return)

Allowed characters: [aA-zZ]+[0-9]+[space]+[@#$%^&+./-""]"""

MODESELECTIONPARTENAIRES_DESCRIPTION = """Selection mode of the partner offers.

This filter works with the partner criterion and is dependent on the originOffer
criterion. Possible values with the results obtained according to the two other filters:

- INCLUS(INCLUDED)
originOffer empty : Returns the PE offers and the Partners listed in the Partners
criterion
originOffer at 2 : Only the offers of the Partners listed in the Partners
criterion
- EXCLU(EXCLUDED)
originOffer empty : Return the offers of PE and Partners not listed in the Partners
criterion
originOffer at 2 : Only the offers of the Partners not listed in the Partners
criterion
Note: In all cases, if originOffer = 1, then only the France Travail (ex: Pole Emploi)
offers will be returned"""


class Experience(str, Enum):
    LESS_THAN_ONE_YEAR = "1"
    ONE_TO_THREE_YEARS = "2"
    MORE_THAN_THREE_YEARS = "3"


class Qualification(IntEnum):
    NON_EXECUTIVE = 0
    EXECUTIVE = 9


class SalaryPeriod(str, Enum):
    MONTHLY = "M"
    ANNUALLY = "A"
    HOURLY = "H"
    FEE_FOR_SERVICE = "C"


class IndustryDomain(str, Enum):
    AGRICULTURE = "A"
    ARTS = "B"
    BANKING_INSURANCE = "C"
    REAL_ESTATE = "C15"
    RETAIL = "D"
    MEDIA_COMMUNICATION = "E"
    CONSTRUCTION = "F"
    HOSPITALITY_TOURISM = "G"
    INDUSTRY = "H"
    INSTALLATION_MAINTENANCE = "I"
    HEALTHCARE = "J"
    PERSONAL_COMMUNITY_SERVICES = "K"
    PERFORMING_ARTS = "L"
    SPORT = "L14"
    PURCHASING_ACCOUNTING_MANAGEMENT = "M"
    EXECUTIVE_MANAGEMENT = "M13"
    CONSULTING_RESEARCH = "M14"
    HUMAN_RESOURCES = "M15"
    SECRETARIAL_ASSISTANTSHIP = "M16"
    MARKETING_SALES_STRATEGY = "M17"
    COMPUTER_TELECOMMUNICATION = "M18"
    TRANSPORTATION_LOGISTICS = "N"


class PublishedSince(IntEnum):
    ONE_DAY = 1
    THREE_DAYS = 3
    SEVEN_DAYS = 7
    FOURTEEN_DAYS = 14
    THIRTY_ONE_DAYS = 31


class PartnerSelectionMode(str, Enum):
    INCLUDE = "INCLUS"
    EXCLUDE = "EXCLU"


class WeeklyDuration(str, Enum):
    NOT_SPECIFIED = "0"
    FULL_TIME = "1"
    PART_TIME = "2"


class AuthParameters(Struct):
    client_id: Annotated[
        str,
        Meta(
            description="Client ID used to access France Travail (ex: Pole Emploi) API",
        ),
    ]
    client_secret: Annotated[
        str,
        Meta(
            description=(
                "Client Secret used to access France Travail (ex: Pole Emploi) API"
            ),
        ),
    ]


class ReadJobsParameters(Struct, omit_defaults=True):
    range: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Pagination of data. The range of results is limited to 150. Format:"
                " p-d, where :\n\np is the index (starting at 0) of the first element"
                " requested, which must not exceed 3000\nd is the index of the last"
                " element requested, which must not exceed 3149"
            ),
        ),
    ] = None
    sort: Annotated[
        t.Optional[int],
        Meta(
            description="Sorting of data",
        ),
    ] = None
    domaine: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Professional field code"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f"{REFERENCES_URL}/domaines"
            ),
        ),
    ] = None
    codeROME: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "ROME code of the profession"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/metiers"
            ),
        ),
    ] = None
    theme: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Theme of the profession"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/themes"
            ),
        ),
    ] = None
    appellation: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "ROME designation code for the offer, see reference below"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/appellations"
            ),
        ),
    ] = None
    codeNAF: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "NAF code of the offer (format 99.99X)"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/nafs"
            ),
        ),
    ] = None
    secteurActivite: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "NAF codes for sectors of activity. It is possible to specify two NAF"
                " codes by separating them with a comma in the character"
                " string.Example : 01,02A GET request for the list of accepted choices"
                " from the Offres d'emploi API to this endpoint :"
                f" {REFERENCES_URL}/secteursActivites"
            ),
        ),
    ] = None
    experience: Annotated[
        t.Optional[Experience],
        Meta(
            description=(
                "Level of experience required\nPossible values:\n1 -> Less than 1 year"
                " of experience\n2 -> From 1 to 3 years of experience\n3 -> More than 3"
                " years of experience"
            ),
        ),
    ] = None
    experienceExigence: Annotated[
        t.Optional[ExperienceRequirement],
        Meta(
            description=(
                "Filter offers by experience requirement (D beginner accepted, S"
                " experience desired, E experience required)"
            ),
        ),
    ] = None
    typeContrat: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Contract type code"
                "Example : CDI,CDD"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/typesContrats"
            ),
        ),
    ] = None
    natureContrat: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the nature of contract"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                " to this endpoint :"
                f" {REFERENCES_URL}/naturesContrats"
            ),
        ),
    ] = None
    origineOffre: Annotated[
        t.Optional[OfferOriginTag],
        Meta(
            description=(
                "Origin of the offer\nPossible values:\n1 -> Job center\n2 -> Partner"
            ),
        ),
    ] = None
    qualification: Annotated[
        t.Optional[Qualification],
        Meta(
            description=(
                "Qualification Code\nPossible values:\n0 -> Non-executive\n9 ->"
                " Executive"
            ),
        ),
    ] = None
    tempsPlein: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Promote the use of the WeeklyDuration filter\nPossible values:\nfalse"
                " -> Part-time\ntrue -> Full time\nIf the parameter is not filled, then"
                " all the offers are returned"
            ),
        ),
    ] = None
    commune: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "INSEE code of the commune"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/communes"
            ),
        ),
    ] = None
    distance: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "Kilometric distance of the search radius\nDefault value: 10Note: to"
                " obtain only the offers of a specific commune, then you must fill in"
                " the parameter 'distance=0'."
            ),
        ),
    ] = None
    # distance=0 pour pour obtenir seulement les offres d'une commune spécifique
    departement: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Job department"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/departements"
            ),
        ),
    ] = None
    inclureLimitrophes: Annotated[
        t.Optional[bool],
        Meta(
            description="Include bordering departments in the search",
        ),
    ] = None
    region: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the region of the offer"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/regions"
            ),
        ),
    ] = None
    paysContinent: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the country or continent of the offer"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/pays"
                f"AND {REFERENCES_URL}/continents"
            ),
        ),
    ] = None
    niveauFormation: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Level of education required"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/niveauxFormations"
            ),
        ),
    ] = None
    permis: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the requested license"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {REFERENCES_URL}/permis"
            ),
        ),
    ] = None
    motsCles: Annotated[
        t.Optional[str],
        Meta(
            description=MOTS_CLES_DESCRIPTION,
        ),
    ] = None
    salaireMin: Annotated[
        t.Optional[float],
        Meta(
            description=(
                "Minimum wage, expressed in Euro.If this data is filled in, the code of"
                " the type of minimum wage is mandatory."
            ),
        ),
    ] = None  # If this field is set, then periodeSalaire is required.
    periodeSalaire: Annotated[
        t.Optional[SalaryPeriod],
        Meta(
            description=(
                "Period for the calculation of the minimum wage.\nIf this data is"
                " filled in, the minimum wage is mandatory.\nPossible values:\nM ->"
                " Monthly\nA -> Annual\nH -> Hourly\nC -> Fee"
            ),
        ),
    ] = None  # If this field is set, then salaireMin is required.
    accesTravailleurHandicape: Annotated[
        t.Optional[bool],
        Meta(
            description="Jobs for which the employer is disabled-friendly",
        ),
    ] = None
    offresMRS: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                " Allows you to search for jobs that offer the simulation recruitment"
                " method"
            ),
        ),
    ] = None
    grandDomaine: Annotated[
        t.Optional[IndustryDomain],
        Meta(
            description="Code of the major area of the offer",
        ),
    ] = None
    experienceExige: Annotated[
        t.Optional[ExperienceRequirement],
        Meta(
            description="Filter offers by experience level.",
        ),
    ] = None
    publieeDepuis: Annotated[
        t.Optional[PublishedSince],
        Meta(
            description=(
                "Maximum number of days since the publication of the offer\nPossible"
                " values: 1, 3, 7, 14, 31"
            ),
        ),
    ] = None
    minCreationDate: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Minimum offer creation date.\nIf this data is filled in, the maximum"
                " offer creation date is mandatory.\nISO-8601 standard"
                " (YYYY-MM-DDTHH:MM:SSZ)"
            ),
        ),
    ] = None  # If this field is set, then minCreationDate is required.
    maxCreationDate: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Maximum offer creation date.\nIf this data is filled in, the minimum"
                " offer creation date is mandatory.\nISO-8601 standard"
                " (YYYY-MM-DDTHH:MM:SSZ)"
            ),
        ),
    ] = None  # If this field is set, then maxCreationDate is required.

    partenaires: Annotated[
        t.Optional[str],
        Meta(
            description=(
                " This filter allows you to enter your partner code in order to include"
                " or exclude your offers from the results according to the"
                " selectionmade in the PartnerSelection mode filter\nIt is possible to"
                " enter several codes (separator ','). "
            ),
        ),
    ] = None  # Il est possible de saisir plusieurs codes (séparateur ",").
    modeSelectionPartenaires: Annotated[
        t.Optional[PartnerSelectionMode],
        Meta(
            description=MODESELECTIONPARTENAIRES_DESCRIPTION,
        ),
    ] = None  # If this field is set, then maxCreationDate is required.

    dureeHebdo: Annotated[
        t.Optional[WeeklyDuration],
        Meta(
            description=(
                "Filtre les offres selon la durée hebdomadaire.\nValeurs possibles :\n0"
                " -> Non précisé\n1 -> Temps plein\n2 -> Temps partiel"
            ),
        ),
    ] = None
    dureeHebdoMin: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "Minimum weekly duration of the offer\nThe value must be in HHMM"
                " format, for example : 8h => 800 ; 24h30 => 2430"
            ),
        ),
    ] = None  # format HHMM
    dureeHebdoMax: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "Maximum weekly duration of the offer\nThe value must be in HHMM"
                " format, for example: 8h => 800; 24h30 => 2430"
            ),
        ),
    ] = None  # format HHMM
    dureeContratMin: Annotated[
        t.Optional[float],
        Meta(
            description=(
                "Minimum duration of the sought contract.\nThe search is done in months"
                " (ex: 0.5 for 15 days, 1.0 for 1 month,2.0 for 2 months).\nPositive"
                " decimal (Decimal separator: '.')"
            ),
        ),
    ] = None
    dureeContratMax: Annotated[
        t.Optional[float],
        Meta(
            description=(
                "Maximum duration of the sought contract.\nThe search is made in months"
                " (ex: 0.5 for 15 days, 1.0 for 1 month,2.0 for 2 months).\nPositive"
                " decimal (Decimal separator: '.')"
            ),
        ),
    ] = None
    offresManqueCandidats: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Filters offers older than 15 days, with less than 4 applications (of"
                " which Pôle emploi is informed)\nfalse -> Offers not concerned\ntrue"
                " -> Offers with few candidates"
            ),
        ),
    ] = None
    entreprisesAdaptees: Annotated[
        t.Optional[bool],
        Meta(
            description=(
                "Filter the offers where the adapted company allows a disabled worker"
                " to exercise a professional activity in  conditions adapted to his"
                " capacities\nfalse -> Offers not concerned\ntrue -> Offers from"
                " adapted companies"
            ),
        ),
    ] = None

    def __post_init__(self):
        # Validate date fields
        validate_date(self.minCreationDate, "minCreationDate")
        validate_date(self.maxCreationDate, "maxCreationDate")


def get_poleemploi_auth_token(client_id: str, client_secret: str) -> str:
    response = requests.post(
        TOKEN_GENERATOR_URL,
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        data=dict(
            grant_type=GRANT_TYPE,
            scope=TOKEN_SCOPE,
            client_id=client_id,
            client_secret=client_secret,
        ),
        params=dict(
            realm="/partenaire",
        ),
    )
    if response.status_code != 200:
        raise ValueError(
            "Failed to get auth token from France Travail (ex: Pole Emploi):"
            f" {response.text}"
        )
    return response.json()["access_token"]


def read(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadJobsParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    token = get_poleemploi_auth_token(
        client_id=auth_parameters.client_id,
        client_secret=auth_parameters.client_secret,
    )
    params = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)

    response = requests.get(
        JOBS_SEARCH_URL,
        headers={"Authorization": "Bearer {}".format(token)},
        params=params,
    )
    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to pull jobs from France Travail (ex: Pole Emploi) params={}"
            " status_code={} response={}".format(
                params, response.status_code, response.text
            )
        )
        raise Exception("Failed to pull jobs from France Travail (ex: Pole Emploi)")
    jobs = response.json()["resultats"]
    for job in jobs:
        yield job


JobsAisle = Aisle(
    name=Entity.job,
    schema=FranceTravailJobOffer,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(create=read, update=read, archive=read),
        endpoints=Endpoints(
            create=SEARCH_JOBS_ENDPOINT,
            update=SEARCH_JOBS_ENDPOINT,
            archive=SEARCH_JOBS_ENDPOINT,
        ),
    ),
)



import datetime
import typing as t
from enum import Enum

import msgspec
from msgspec import Struct


class Category(Struct):
    tag: str
    label: str


class Location(msgspec.Struct):
    area: t.List[str]
    display_name: str


class ContractType(str, Enum):
    PERMANENT = "permanent"
    CONTRACT = "contract"


class Flag(str, Enum):
    YES = "1"
    NO = "0"


class ContractTime(str, Enum):
    FULL_TIME = "full_time"
    PART_TIME = "part_time"


class Company(msgspec.Struct):
    display_name: str
    canonical_name: t.Optional[str]
    count: t.Optional[int]


class AdzunaJob(Struct):
    id: str
    created: str
    title: str
    description: str
    full_description: t.Optional[str]
    redirect_url: str
    latitude: t.Optional[float]
    longitude: t.Optional[float]
    category: Category
    location: Location
    salary_min: int
    salary_max: int
    salary_is_predicted: Flag
    company: Company
    contract_type: t.Optional[ContractType]
    contract_time: t.Optional[ContractTime]

    def __post_init__(self):
        self._validate_created()

    def _validate_created(self):
        try:
            datetime.date.fromisoformat(self.created)
        except ValueError:
            raise ValueError(
                f"Invalid time format given '{self.created}', expected ISO 8601 format"
                " (YYYY-MM-DDTHH:MM:SSZ)"
            )



import typing as t
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec import json as msgspec_json
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.taleez.schemas import (
    Candidate,
    ContractType,
    Job,
    JobStatus,
    JobVisibility,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

TALEEZ_PROFILES_ENDPOINT = "https://api.taleez.com/0/candidates"
TALEEZ_JOBS_ENDPOINT = "https://api.taleez.com/0/jobs"
TALEEZ_JOBS_ENDPOINT_LIMIT = 100
ACCEPT = "application/json;charset=UTF-8"
CONTENT_TYPE = "application/json"


class AuthParameters(Struct):
    x_taleez_api_secret: Annotated[
        str,
        Meta(
            description="X-taleez-api-secret used to access Taleez API",
        ),
    ]


class ReadJobsParameters(Struct, omit_defaults=True):
    unitId: Annotated[
        t.Optional[t.List[int]],
        Meta(
            description="Filter on job unit. Can be used with others filters.",
        ),
    ] = None
    status: Annotated[
        t.Optional[t.List[JobStatus]],
        Meta(
            description=(
                "Filter on job status. Can be used with others filters.\nAvailable"
                " values : {}".format([e.value for e in JobStatus])
            ),
        ),
    ] = None
    contract: Annotated[
        t.Optional[t.List[ContractType]],
        Meta(
            description="Filter on job contract. Can be used with others filters.",
        ),
    ] = None
    city: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="Filter on job city. Can be used with others filters.",
        ),
    ] = None
    companyLabel: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description=(
                "Filter on job company label (strict search). Can be used with others"
                " filters."
            ),
        ),
    ] = None
    tag: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="Filter on job tag. Can be used with others filters.",
        ),
    ] = None
    visibility: Annotated[
        t.Optional[t.List[JobVisibility]],
        Meta(
            description="Filter on job visibility. Can be used with others filters.",
        ),
    ] = None
    visibilityToken: Annotated[
        t.Optional[str],
        Meta(
            description="Secret token for restricted jobs.",
        ),
    ] = None
    sort: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Sort the list by one or multiple params. Ex :"
                " sort=dateCreation.desc,label.asc"
            ),
        ),
    ] = None


class ReadProfilesParameters(Struct, omit_defaults=True):
    mail: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="Filter by mail",
        ),
    ] = None


class WriteProfilesParameters(Struct):
    recruiterId: Annotated[
        t.Optional[int],
        Meta(
            description="Id of the recruiter to associate the candidate with",
        ),
    ] = None
    unitId: Annotated[
        t.Optional[int],
        Meta(
            description=(
                "Id of the unit associated with the candidate. Only specified for"
                ' companies with multiple units and "candidate segmentation by unit"'
                " setting."
            )
        ),
    ] = None
    job_ids: Annotated[
        t.Optional[t.List[int]],
        Meta(
            description="List of job ids to associate the candidate with",
        ),
    ] = None
    pool_ids: Annotated[
        t.Optional[t.List[int]],
        Meta(
            description="List of pool ids to associate the candidate with",
        ),
    ] = None


class UpdateProfilesParameters(Struct):
    pass


def read_jobs(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadJobsParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    params = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)
    params["withDetails"] = True
    params["withProps"] = True
    page = 0
    jobs = []

    while True:
        params["page"] = page
        response = requests.get(
            TALEEZ_JOBS_ENDPOINT,
            headers={"X-taleez-api-secret": auth_parameters.x_taleez_api_secret},
            params=params,
        )

        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Taleez params={} status_code={} response={}"
                .format(params, response.status_code, response.text)
            )
            raise Exception("Failed to pull jobs from Taleez")

        response = response.json()
        jobs.extend(response["list"])
        if response["hasMore"] is False:
            break
        page += 1

    for job in jobs:
        yield job


def read_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    params = msgspec_json.decode(msgspec_json.encode(parameters), type=dict)
    params["withProps"] = True
    page = 0
    profiles = []

    while True:
        params["page"] = page
        response = requests.get(
            TALEEZ_PROFILES_ENDPOINT,
            headers={"X-taleez-api-secret": auth_parameters.x_taleez_api_secret},
            params=params,
        )

        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from Taleez params={} status_code={}"
                " response={}".format(params, response.status_code, response.text)
            )
            raise Exception("Failed to pull profiles from Taleez")

        response = response.json()
        profiles.extend(response["list"])
        if response["hasMore"] is False:
            break
        page += 1

    for profile in profiles:
        resume_link = profile.get("resume")
        if resume_link:
            resume = requests.get(resume_link)
            if resume.status_code // 100 == 2:
                profile["resume"] = resume.content
        yield profile


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    items = list(items)
    adapter.info("Pushing {} candidates to Taleez API".format(len(items)))
    failed_profiles = []

    for profile_item in items:
        candidate = profile_item.pop("candidate")
        resume_link = profile_item.pop("resume")
        properties = profile_item.pop("properties", None)

        post_profile_response = requests.post(
            TALEEZ_PROFILES_ENDPOINT,
            headers={"X-taleez-api-secret": auth_parameters.x_taleez_api_secret},
            json=candidate,
        )
        if post_profile_response.status_code // 100 != 2:
            adapter.error(
                "Failed to create candidate status_code={} post_profile_response={}"
                .format(
                    post_profile_response.status_code,
                    post_profile_response.text,
                )
            )
            failed_profiles.append(profile_item)
            continue

        id = post_profile_response.json()["id"]
        binary_resume = requests.get(resume_link).content
        post_resume_response = requests.post(
            "{}/{}/documents".format(TALEEZ_PROFILES_ENDPOINT, id),
            headers={
                "X-taleez-api-secret": auth_parameters.x_taleez_api_secret,
            },
            params={"cv": True},
            files={
                "file": (
                    "resume_{}_{}".format(
                        candidate["firstName"],
                        candidate["lastName"],
                    ),
                    binary_resume,
                    "application/pdf",
                )
            },
        )
        if post_resume_response.status_code // 100 != 2:
            adapter.error(
                "Failed to add resume status_code={} post_resume_response={}".format(
                    post_resume_response.status_code,
                    post_resume_response.text,
                )
            )

        if properties:
            post_properties_response = requests.post(
                "{}/{}/properties".format(TALEEZ_PROFILES_ENDPOINT, id),
                headers={
                    "X-taleez-api-secret": auth_parameters.x_taleez_api_secret,
                },
                json=properties,
            )
            if post_properties_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to update property status_code={}"
                    " post_properties_response={}".format(
                        post_properties_response.status_code,
                        post_properties_response.text,
                    )
                )

        if parameters.job_ids:
            associate_job_response = requests.post(
                "{}/{}/jobs".format(TALEEZ_PROFILES_ENDPOINT, id),
                headers={
                    "X-taleez-api-secret": auth_parameters.x_taleez_api_secret,
                },
                json={"ids": parameters.job_ids},
            )
            if associate_job_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to associate candidate to jobs status_code={}"
                    " associate_job_response={}".format(
                        associate_job_response.status_code, associate_job_response.text
                    )
                )
        if parameters.pool_ids:
            associate_pool_response = requests.post(
                "{}/{}/pools".format(TALEEZ_PROFILES_ENDPOINT, id),
                headers={
                    "X-taleez-api-secret": auth_parameters.x_taleez_api_secret,
                },
                json={"ids": parameters.pool_ids},
            )
            if associate_pool_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to associate candidate to pools status_code={}"
                    " associate_pool_response={}".format(
                        associate_pool_response.status_code,
                        associate_pool_response.text,
                    )
                )

    return failed_profiles


def update_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    for profile_item in items:
        update_profile_response = requests.post(
            "{}/{}".format(TALEEZ_PROFILES_ENDPOINT, profile_item["id"]),
            headers={"X-taleez-api-secret": auth_parameters.x_taleez_api_secret},
            json=profile_item["candidate"],
        )
        if update_profile_response.status_code // 100 != 2:
            adapter.error(
                "Failed to update candidate status_code={} response={}".format(
                    update_profile_response.status_code,
                    update_profile_response.text,
                )
            )
            failed_profiles.append(profile_item)
            continue

        if profile_item.get("properties"):
            update_properties_response = requests.post(
                "{}/{}/properties".format(TALEEZ_PROFILES_ENDPOINT, profile_item["id"]),
                headers={"X-taleez-api-secret": auth_parameters.x_taleez_api_secret},
                json=profile_item["properties"],
            )
            if update_properties_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to update properties status_code={} response={}".format(
                        update_properties_response.status_code,
                        update_properties_response.text,
                    )
                )

    return failed_profiles


ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=Candidate,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=read_profiles,
            update=read_profiles,
            archive=read_profiles,
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=UpdateProfilesParameters,
        ),
        function=merge(
            create=write_profiles,
            update=update_profiles,
        ),
    ),
)
JobsAisle = Aisle(
    name=Entity.job,
    schema=Job,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=read_jobs,
            update=read_jobs,
            archive=read_jobs,
        ),
    ),
)



import collections
import collections.abc
import typing as t
from dataclasses import dataclass
from functools import reduce

from msgspec import Struct
from msgspec import ValidationError as MsgSpecValidationError
from msgspec import convert, json
from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError

from hrflow_connectors.v2.core.common import Parameters, Schema

T = t.TypeVar("T", bound=t.Union[Struct, BaseModel])


class ValidationError(Exception):
    pass


@t.overload
def serialize(obj: dict, schema: type[T]) -> T:
    ...  # pragma: nocover


@t.overload
def serialize(obj: dict, schema: Schema) -> Parameters:
    ...  # pragma: nocover


def serialize(obj: dict, schema: Schema) -> Parameters:
    if issubclass(schema, BaseModel):
        try:
            return schema(**obj)
        except PydanticValidationError as e:
            raise ValidationError(e.errors())
    try:
        return convert(obj, schema)
    except MsgSpecValidationError as e:
        raise ValidationError(*e.args)


def fields(schema: Schema) -> tuple[str, ...]:
    if issubclass(schema, BaseModel):
        return tuple(schema.__fields__)
    return schema.__struct_fields__


def msgspec_schema_hook(t: type):
    if issubclass(t, collections.abc.Callable):
        return dict(type="Callable")
    raise NotImplementedError  # pragma: nocover


def json_schema(schema: Schema, unwrap: bool = True) -> dict:
    if issubclass(schema, BaseModel):
        return schema.schema()

    wrapped = json.schema(schema, schema_hook=msgspec_schema_hook)
    if unwrap:
        path = wrapped.pop("$ref").rsplit("/", 1)[-1]
        unwrapped = wrapped["$defs"].pop(path)
        return {**unwrapped, "$defs": wrapped["$defs"]}
    return wrapped


@dataclass
class TemplateField:
    name: str
    type: str
    required: bool
    description: str
    default: str


def get_type(definition: dict, json_schema: dict) -> str:
    if type := definition.get("type"):
        return type

    if (
        sub_definitions := definition.get("anyOf", definition.get("allOf"))
    ) is not None:
        return "|".join(
            [
                get_type(sub_definition, json_schema)
                for sub_definition in t.cast(list[dict], sub_definitions)
            ]
        )
    if ref := definition.get("$ref"):
        return get_type(
            reduce(
                lambda reduced, path: reduced[path],
                ref.strip("#/").split("/"),
                json_schema,
            ),
            json_schema,
        )
    if choices := definition.get("enum"):
        return "Literal[" + ",".join([f"'{choice}'" for choice in choices]) + "]"

    return ""  # pragma: nocover


def template_fields(schema: Schema) -> list[TemplateField]:
    _json_schema = json_schema(schema)
    return [
        TemplateField(
            name=key,
            type=get_type(definition, _json_schema),
            required=key in _json_schema["required"],
            description=definition.get("description", ""),
            default=definition.get("default", None),
        )
        for key, definition in _json_schema["properties"].items()
    ]



from hrflow_connectors.v2.core.hrflow.aisles import (
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

HrFlowWarehouse = Warehouse(auth=AuthParameters, aisles=(JobsAisle, ProfilesAisle))



from typing import Optional, Union

from msgspec import Meta, Struct
from typing_extensions import Annotated


class Location(Struct):
    location_str: Optional[str]
    country: Optional[str]
    country_code: Optional[str]
    region: Optional[str]
    region_code: Optional[str]
    city: Optional[str]
    zip_code: Optional[str]
    telecommuting: Optional[bool]


class WorkableJob(Struct):
    id: Optional[str]
    title: str
    full_title: str
    shortcode: str
    code: Optional[str]
    state: Optional[str]
    department: Optional[str]
    url: Optional[str]
    application_url: Optional[str]
    shortlink: Optional[str]
    location: Location
    created_at: str
    description: Optional[str]
    requirements: Optional[str]
    benefit: Optional[str]
    employment_type: Optional[str]


class EducationEntry(Struct):
    school: str
    degree: Optional[str]
    field_of_study: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]


class ExperienceEntry(Struct):
    company: Optional[str]
    title: str
    summary: Optional[str]
    industry: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]
    current: Optional[bool]


class SocialProfile(Struct):
    type: str
    name: str
    username: str
    url: str


class WorkableCandidate(Struct):
    name: str
    firstname: str
    lastname: str
    email: str
    headline: Annotated[
        Optional[str],
        Meta(description="One line description as provided by the candidate or you"),
    ]
    summary: Annotated[
        Optional[str], Meta(description="The profile summary provided by the candidate")
    ]
    address: Optional[str]
    phone: Optional[str]
    cover_letter: Optional[str]
    education_entries: Optional[list[EducationEntry]]
    experience_entries: Optional[list[ExperienceEntry]]
    answers: Optional[list[dict]]
    skills: Optional[Union[str, list[str]]]
    tags: Optional[list[str]]
    disqualified: Optional[bool]
    disqualification_reason: Optional[str]
    disqualified_at: Optional[str]
    social_profiles: Optional[list[SocialProfile]]
    domain: Optional[str]
    resume_url: Optional[str]
    recruiter_key: Optional[str]



# This file is generated automatically
from hrflow_connectors.v2.core.connector import Connector, PublicActionInterface

class AdzunaProto(Connector):
    create_jobs_in_hrflow: PublicActionInterface
    update_jobs_in_hrflow: PublicActionInterface
    archive_jobs_in_hrflow: PublicActionInterface
    
Adzuna: AdzunaProto


import time
import typing as t
from collections import Counter
from datetime import datetime, timezone
from enum import Enum
from logging import Logger, LoggerAdapter, getLogger
from uuid import uuid4

from msgspec import Struct, field

from hrflow_connectors.core import backend
from hrflow_connectors.v2.core.common import Mode, Parameters, Schema
from hrflow_connectors.v2.core.msgspec_pydantic_compat import ValidationError, serialize
from hrflow_connectors.v2.core.warehouse import Aisle

default_logger = getLogger(__name__)


class LogTag(Struct):
    name: str
    value: str


def get_adapter(tags: t.Sequence[LogTag], logger: t.Optional[Logger] = None):
    formatted_tags = "".join(["[{}={}]".format(tag.name, tag.value) for tag in tags])

    class Adapter(LoggerAdapter):
        def process(self, msg: str, kwargs: t.Any) -> tuple[str, dict]:
            return (
                "{}: {}".format(
                    formatted_tags,
                    msg,
                ),
                kwargs,
            )

    return Adapter(
        logger=logger or default_logger,
        extra=dict(tags={tag.name: tag.value for tag in tags}),
    )


class Event(Enum):
    read_success = "read_success"
    read_failure = "read_failure"
    format_failure = "format_failure"
    logics_discard = "logics_discard"
    logics_failure = "logics_failure"
    write_failure = "write_failure"
    callback_failure = "callback_failure"
    callback_executed = "callback_executed"
    getting_incremental_token_failure = "getting_incremental_token_failure"

    @classmethod
    def empty_counter(cls) -> t.Counter["Event"]:
        return Counter({event: 0 for event in cls})


class Reason(Enum):
    workflow_id_not_found = "workflow_id_not_found"
    origin_is_not_readable = "origin_is_not_readable"
    target_is_not_writable = "target_is_not_writable"
    origin_does_not_support_incremental = "origin_does_not_support_incremental"
    backend_not_configured = "backend_not_configured"
    event_parsing_failure = "event_parsing_failure"
    getting_incremental_token_failure = "getting_incremental_token_failure"
    mode_not_supported_by_origin = "mode_not_supported_by_origin"
    mode_not_supported_by_target = "mode_not_supported_by_target"
    bad_origin_parameters = "bad_origin_parameters"
    bad_target_parameters = "bad_target_parameters"
    format_failure = "format_failure"
    logics_failure = "logics_failure"
    read_failure = "read_failure"
    write_failure = "write_failure"
    none = "none"


class Status(Enum):
    success = "success"
    success_with_failures = "success_with_failures"
    fatal = "fatal"


class ActionInitError(Struct):
    data: dict
    reason: Reason


class RunResult(Struct):
    incremental: bool
    status: Status
    reason: Reason = Reason.none
    events: t.Counter[Event] = field(default_factory=Event.empty_counter)
    incremental_token: t.Optional[str] = None

    @classmethod
    def from_events(
        cls,
        events: t.Counter[Event],
        incremental: bool,
        incremental_token: t.Optional[str] = None,
    ) -> "RunResult":
        read_success = events[Event.read_success]
        read_failures = events[Event.read_failure]
        if read_success == 0 and read_failures == 0:
            return cls(
                status=Status.success,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )
        elif read_success == 0 and read_failures > 0:
            return cls(
                status=Status.fatal,
                reason=Reason.read_failure,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

        logics_failures = events[Event.logics_failure]
        if logics_failures == read_success:
            return cls(
                status=Status.fatal,
                reason=Reason.logics_failure,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

        logics_discard = events[Event.logics_discard]
        format_failures = events[Event.format_failure]
        if format_failures == read_success - logics_failures - logics_discard:
            return cls(
                status=Status.fatal,
                reason=Reason.format_failure,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

        write_failure = events[Event.write_failure]
        if (
            write_failure
            == read_success - logics_discard - logics_failures - format_failures
        ) and write_failure > 0:
            return cls(
                status=Status.fatal,
                reason=Reason.write_failure,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

        has_failures = any(
            events[event] > 0
            for event in [
                Event.read_failure,
                Event.format_failure,
                Event.logics_failure,
                Event.write_failure,
                Event.callback_failure,
            ]
        )
        if has_failures:
            return cls(
                status=Status.success_with_failures,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )
        return cls(
            status=Status.success,
            events=events,
            incremental=incremental,
            incremental_token=incremental_token,
        )


class Metadata(Struct):
    connector_name: str
    origin_name: str
    target_name: str
    action_name: str
    using_default_format: bool
    using_default_logics: bool


FormatT = t.Callable[[dict], dict]
LogicsT = list[t.Callable[[dict], t.Optional[dict]]]


class CallbackT(t.Protocol):
    def __call__(
        self,
        origin_parameters: Parameters,
        target_parameters: Parameters,
        events: Counter[Event],
        items: list[dict],
    ) -> None:
        ...  # pragma: nocover


def run(
    *,
    workflow_id: str,
    metadata: Metadata,
    mode: Mode,
    origin: Aisle,
    origin_auth_schema: Schema,
    origin_auth: dict,
    origin_parameters: dict,
    target: Aisle,
    target_auth_schema: Schema,
    target_auth: dict,
    target_parameters: dict,
    init_error: t.Optional[ActionInitError] = None,
    incremental: bool = False,
    format: t.Optional[FormatT] = None,
    logics: t.Optional[LogicsT] = None,
    callback: t.Optional[CallbackT] = None,
    persist: bool = True,
):
    action_id = uuid4()
    started_at = datetime.now(tz=timezone.utc)
    adapter = get_adapter(
        tags=[
            LogTag(name="workflow_id", value=workflow_id),
            LogTag(name="action_id", value=action_id.hex),
            LogTag(name="connector", value=metadata.connector_name),
            LogTag(name="origin", value=metadata.origin_name),
            LogTag(name="target", value=metadata.target_name),
            LogTag(name="action_name", value=metadata.action_name),
            LogTag(name="started_at", value=started_at.isoformat()),
        ]
    )

    if init_error is not None:
        adapter.error(
            "Failed to parse event with reason={} data={}".format(
                repr(init_error.reason), init_error.data
            )
        )
        return RunResult(
            status=Status.fatal, reason=init_error.reason, incremental=incremental
        )

    if origin.read is None:
        return RunResult(
            status=Status.fatal,
            reason=Reason.origin_is_not_readable,
            incremental=incremental,
        )

    if target.write is None:
        return RunResult(
            status=Status.fatal,
            reason=Reason.target_is_not_writable,
            incremental=incremental,
        )

    origin_parameters_schema = origin.parameters(operation="read", mode=mode)
    if origin_parameters_schema is None:
        return RunResult(
            status=Status.fatal,
            reason=Reason.mode_not_supported_by_origin,
            incremental=incremental,
        )

    target_parameters_schema = target.parameters(operation="write", mode=mode)
    if target_parameters_schema is None:
        return RunResult(
            status=Status.fatal,
            reason=Reason.mode_not_supported_by_target,
            incremental=incremental,
        )

    adapter.info("Starting Action")
    try:
        parsed_origin_auth_parameters = serialize(origin_auth, origin_auth_schema)
    except ValidationError as e:
        adapter.warning(f"Failed to parse origin_auth with errors={e}")
        return RunResult(
            status=Status.fatal,
            reason=Reason.bad_origin_parameters,
            incremental=incremental,
        )

    try:
        parsed_origin_parameters = serialize(
            origin_parameters, origin_parameters_schema
        )
    except ValidationError as e:
        adapter.warning(f"Failed to parse origin_parameters with errors={e}")
        return RunResult(
            status=Status.fatal,
            reason=Reason.bad_origin_parameters,
            incremental=incremental,
        )

    try:
        parsed_target_auth_parameters = serialize(target_auth, target_auth_schema)
    except ValidationError as e:
        adapter.warning(f"Failed to parse target_auth with errors={e}")
        return RunResult(
            status=Status.fatal,
            reason=Reason.bad_target_parameters,
            incremental=incremental,
        )

    try:
        parsed_target_parameters = serialize(
            target_parameters, target_parameters_schema
        )
    except ValidationError as e:
        adapter.warning(f"Failed to parse target_parameters with errors={e}")
        return RunResult(
            status=Status.fatal,
            reason=Reason.bad_target_parameters,
            incremental=incremental,
        )

    incremental_token = None
    if incremental:
        if origin.read.supports_incremental(mode) is False:
            adapter.warning(
                f"Origin warehouse {metadata.origin_name} does not support incremetal"
                " reading for mode {mode}."
            )
            return RunResult(
                status=Status.fatal,
                reason=Reason.origin_does_not_support_incremental,
                incremental=incremental,
            )

        if backend.store is None:
            adapter.warning("Backend not configured: Cannot run in incremental mode")
            return RunResult(
                status=Status.fatal,
                reason=Reason.backend_not_configured,
                incremental=incremental,
            )

        adapter.info("Reading in incremental mode: fetching last token")
        last_results = backend.store.load(key=workflow_id, parse_as=RunResult)
        incremental_token = (
            last_results.incremental_token if last_results is not None else None
        )

    events = Event.empty_counter()

    read_started_at = time.time()
    adapter.info(
        f"Starting to read from warehouse={metadata.origin_name} with "
        f"mode={mode} incremental={incremental} incremental_token={incremental_token}"
    )
    origin_items = []
    try:
        for item in origin.read(
            mode=mode,
            adapter=adapter,
            auth_parameters=parsed_origin_auth_parameters,
            parameters=parsed_origin_parameters,
            incremental=incremental,
            incremental_token=incremental_token,
        ):
            origin_items.append(item)
            events[Event.read_success] += 1
    except Exception as e:
        events[Event.read_failure] += 1
        adapter.exception(
            f"Failed to read from warehouse={metadata.origin_name} with error={repr(e)}"
        )
    if len(origin_items) == 0:
        if events[Event.read_failure] > 0:
            adapter.warning(
                "No items fetched from origin warehoue. Aborting action after"
                " read_failure"
            )
        return RunResult.from_events(
            events, incremental=incremental, incremental_token=incremental_token
        )

    read_finished_at = time.time()
    adapter.info(
        f"Finished reading in {read_finished_at - read_started_at} from"
        f" warehouse={metadata.origin_name}"
        f" n_items={len(origin_items)} read_failure={events[Event.read_failure]}"
    )

    next_incremental_token = incremental_token
    if len(origin_items) > 0 and incremental:
        last_item = origin_items[-1]

        # We know it's not None because of the check
        # origin.read.supports_incremental(mode)
        # Adding these kinds of asserts which are anyway removed
        # in optimized Python bytecode is for type checkers only
        assert origin.read.incremental_token_handler is not None
        token_handler = token_handler = origin.read.incremental_token_handler(mode)
        assert token_handler is not None  # for type-checker

        try:
            next_incremental_token = token_handler(last_item)
        except Exception as e:
            events[Event.getting_incremental_token_failure] += 1
            adapter.exception(
                f"Failed to get read_from from warehouse={metadata.origin_name} with"
                f" error={repr(e)}"
            )
            return RunResult(
                status=Status.fatal,
                reason=Reason.getting_incremental_token_failure,
                events=events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

    if logics is None or len(logics) == 0:
        if logics is None:
            adapter.info("No logics supplied: Skipping ...")
        elif len(logics) == 0:
            adapter.info("Empty logics array supplied: Skipping ...")
        selected_items = origin_items
    else:
        adapter.info(
            "Starting to apply logic functions: "
            f"n_items={len(origin_items)} before applying logics"
        )
        selected_items = []
        for item in origin_items:
            for i, logic in enumerate(logics):
                try:
                    item = logic(item)
                except Exception as e:
                    adapter.exception(
                        f"Failed to apply logic function number={i} error={repr(e)}"
                    )
                    events[Event.logics_failure] += 1
                    break
                if item is None:
                    events[Event.logics_discard] += 1
                    break
            else:
                selected_items.append(item)

        if len(selected_items) == 0:
            adapter.warning(
                "Logics failed for all items. Review supplied logic functions."
                " Aborting action."
            )
            return RunResult.from_events(
                events,
                incremental=incremental,
                incremental_token=incremental_token,
            )
        adapter.info(
            "Finished applying logic functions: "
            f"success={len(selected_items)} discarded={events[Event.logics_discard]}"
            f" failures={events[Event.logics_failure]}"
        )

    if format is None:
        adapter.info("No format function supplied: Skipping ...")
        formatted_items = selected_items
    else:
        adapter.info(
            "Starting to format origin items using"
            f" {'default' if metadata.using_default_format else 'user defined'} function"
        )
        formatted_items = []
        for item in selected_items:
            try:
                formatted_items.append(format(item))
            except Exception as e:
                events[Event.format_failure] += 1
                adapter.exception(
                    "Failed to format origin item using"
                    f" {'default' if metadata.using_default_format else 'user defined'}"
                    f" function error={repr(e)}"
                )
        adapter.info(
            "Finished formatting origin items"
            f" success={len(formatted_items)} failures={events[Event.format_failure]}"
        )

        if len(formatted_items) == 0:
            adapter.warning(
                "Formatting failed for all items. Review supplied format function."
                " Aborting action."
            )
            return RunResult.from_events(
                events,
                incremental=incremental,
                incremental_token=incremental_token,
            )

    if persist is False:
        adapter.info(
            f"Running in dry mode with persist={persist}: Ending execution after read,"
            " format and logics"
        )
        return RunResult.from_events(
            events, incremental=incremental, incremental_token=incremental_token
        )

    write_started_at = time.time()
    adapter.info(
        f"Starting to write to warehouse={metadata.target_name} with"
        f" n_items={len(formatted_items)}"
    )
    try:
        failed_items = target.write(
            mode=mode,
            adapter=adapter,
            auth_parameters=parsed_target_auth_parameters,
            parameters=parsed_target_parameters,
            items=formatted_items,
        )
        events[Event.write_failure] += len(failed_items)
    except Exception as e:
        adapter.exception(
            f"Failed to write to warehouse={metadata.target_name} with error={repr(e)}"
        )
        events[Event.write_failure] += len(formatted_items)
        return RunResult(
            status=Status.fatal,
            reason=Reason.write_failure,
            events=events,
            incremental=incremental,
            incremental_token=incremental_token,
        )
    write_finished_at = time.time()
    adapter.info(
        f"Finished writing in {write_finished_at - write_started_at} "
        f"to warehouse={metadata.target_name} "
        f"success={len(formatted_items) - events[Event.write_failure]} "
        f"failures={events[Event.write_failure]}"
    )

    if callback is not None:
        adapter.info("Calling callback function")
        try:
            callback(
                parsed_origin_parameters,
                parsed_target_parameters,
                events,
                formatted_items,
            )
        except Exception as e:
            events[Event.callback_failure] += 1
            adapter.exception(f"Failed to run callback with error={repr(e)}")
        finally:
            events[Event.callback_executed] += 1

    results = RunResult.from_events(
        events, incremental=incremental, incremental_token=next_incremental_token
    )
    if backend.store is not None:
        adapter.info(f"Saving run results in {backend.store.name} backend")
        backend.store.save(key=workflow_id, data=results)

    adapter.info("Finished action")
    return results



from hrflow_connectors.v2.connectors.bullhorn.aisles import (
    ApplicationsAisle,
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

BullhornWarehouse = Warehouse(
    auth=AuthParameters, aisles=(JobsAisle, ProfilesAisle, ApplicationsAisle)
)



from enum import Enum
from typing import Optional

from msgspec import Struct


class FlatchrProfile(Struct):
    applicant: str
    vacancy: str
    column: str
    vacancy_id: int
    external_id: str
    status: int
    score: int
    hired: bool
    firstname: str
    lastname: str
    email: str
    phone: str
    created_at: str
    source: str


class Experience(int, Enum):
    NONE = 1  # ( < 1 year )
    MEDIUM = 2  # ( between 1 and 3 years )
    SENIOR = 4  # ( between 3 and 5 years )
    EXECUTIVE = 6  # ( > 6 years )


class FlatchrVacancy(Struct):
    id: str
    vacancy_id: Optional[int]
    slug: Optional[str]
    reference: Optional[str]
    title: str
    description: str
    experience: Experience
    mission: str
    profile: str
    salary: Optional[int]
    salary_max: Optional[int]
    status: int
    language: str
    contract_type_id: int
    education_level_id: int
    activity_id: int
    channel_id: int
    metier_id: int
    company_id: int
    mensuality: Optional[str]
    apply_url: Optional[str]
    currency: Optional[str]
    created_at: Optional[str]
    updated_at: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]
    driver_license: Optional[bool]
    remote: Optional[str]
    handicap: Optional[bool]
    partial: Optional[bool]
    meta_title: Optional[str]
    meta_description: Optional[str]
    meta_tags: Optional[str]
    options: dict
    video_url: Optional[str]
    address: str
    show_address: Optional[bool]
    show_contract_date: Optional[bool]
    show_contract_type: Optional[bool]
    show_salary: Optional[bool]
    worker_status: Optional[str]
    skills: Optional[str]
    kanban: Optional[bool]
    slug_mail: Optional[str]



import typing as t
from logging import LoggerAdapter

from hrflow import Hrflow
from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.hrflow.aisles.common import AuthParameters
from hrflow_connectors.v2.core.hrflow.schemas import HrFlowProfile
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)


class ReadAllModesCriterias(Struct):
    source_key: Annotated[str, Meta(description="HrFlow.ai source key")]
    profile_key: Annotated[
        str,
        Meta(
            description="HrFlow.ai profile key",
        ),
    ]


class CreateCriterias(Struct):
    source_key: Annotated[
        str,
        Meta(
            description="HrFlow.ai source key",
        ),
    ]


class UpdateCriterias(Struct):
    source_key: Annotated[
        str,
        Meta(
            description="HrFlow.ai source key",
        ),
    ]
    only_edit_fields: Annotated[
        t.Optional[list[str]],
        Meta(
            description=(
                "List of attributes to use for the edit operation e.g. ['tags',"
                " 'metadatas']"
            ),
        ),
    ] = None


class ArchiveCriterias(Struct):
    source_key: Annotated[
        str,
        Meta(
            description="HrFlow.ai source key",
        ),
    ]


def read(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadAllModesCriterias,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> list[dict]:
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    response = hrflow_client.profile.storing.get(
        source_key=parameters.source_key, key=parameters.profile_key
    )
    if "Unable to find object" in response["message"]:
        adapter.info(
            "No profile found for source_key={} profile_key={} response={}".format(
                parameters.source_key, parameters.profile_key, response
            )
        )
        return []
    elif response["code"] >= 400:
        adapter.error(
            "Failed to get profile source_key={} profile_key={} response={}".format(
                parameters.source_key, parameters.profile_key, response
            )
        )
        raise Exception("Failed to get profile")
    return [response["data"]]


def merge_info(base: dict, info: dict) -> dict:
    if not info:
        return base

    info_parsed = base.get("info", {})
    existing_urls = info_parsed.get("urls", [])

    if isinstance(info.get("urls"), list):
        for new_url in info["urls"]:
            if new_url not in existing_urls:
                existing_urls.append(new_url)

    info_parsed["urls"] = existing_urls

    for key, value in info.items():
        if value and key != "location" and key != "urls":
            info_parsed[key] = value
        elif key == "location" and isinstance(value, dict) and any(value.values()):
            info_parsed[key] = value

    base["info"] = info_parsed
    return base


def merge_item(base: dict, profile: dict, item: str) -> dict:
    if not profile.get(item):
        return base

    base[item] = profile[item]
    return base


def hydrate_profile(profile_parsed: dict, profile_json: dict) -> dict:
    profile_info = profile_json.get("info", {})
    profile_enriched = merge_info(profile_parsed, profile_info)

    items_to_merge = [
        "experiences",
        "educations",
        "skills",
        "languages",
        "certifications",
        "interests",
        "tags",
        "metadatas",
    ]
    for item in items_to_merge:
        profile_enriched = merge_item(profile_enriched, profile_json, item)

    profile_enriched["text"] = profile_json.get("text") or profile_enriched.get("text")
    profile_enriched["text_language"] = profile_json.get(
        "text_language"
    ) or profile_enriched.get("text_language")
    profile_enriched["experiences_duration"] = (
        profile_json.get("experiences_duration")
        if profile_json.get("experiences_duration") is not None
        else profile_enriched.get("experiences_duration")
    )
    profile_enriched["educations_duration"] = (
        profile_json.get("educations_duration")
        if profile_json.get("educations_duration") is not None
        else profile_enriched.get("educations_duration")
    )
    return profile_enriched


def create(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: CreateCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    source_response = hrflow_client.source.get(key=parameters.source_key)

    if source_response["code"] != 200:
        adapter.warning(
            "Failed to get source with"
            f" key={parameters.source_key} response={source_response}"
        )
        return failed

    for profile in items:
        if hrflow_client.profile.storing.get(
            source_key=parameters.source_key,
            reference=profile["reference"],
        ).get("data"):
            adapter.info(
                f"Can't create Profile with reference={profile['reference']} already"
                " exists"
            )
            continue

        if not profile.get("resume") or profile["resume"].get("raw") is None:
            adapter.info(
                f"Profile with reference {profile['reference']} has no resume, adding"
                " without parsing"
            )
            response = hrflow_client.profile.storing.add_json(
                source_key=parameters.source_key, profile_json=profile
            )
            if response["code"] // 100 != 2:
                adapter.error(
                    "Failed to add profile without parsing with"
                    f" reference={profile['reference']} response={response}"
                )
                failed.append(profile)
            continue
        else:
            parsing_response = hrflow_client.profile.parsing.add_file(
                source_key=parameters.source_key,
                profile_file=profile["resume"]["raw"],
                profile_content_type=profile["resume"].get("content_type"),
                profile_file_name=profile["resume"].get("file_name"),
                reference=profile["reference"],
                tags=profile.get("tags", []),
                metadatas=profile.get("metadatas", {}),
                created_at=profile.get("created_at"),
            )

            if parsing_response["code"] not in [202, 201]:
                adapter.error(
                    "Failed to parse profile with"
                    f" reference={profile['reference']} response={parsing_response}"
                )
                failed.append(profile)
                continue

        if source_response["data"].get("sync_parsing"):
            current_profile = parsing_response["data"]["profile"]
            profile_result = hydrate_profile(current_profile, profile)
            response = hrflow_client.profile.storing.edit(
                source_key=parameters.source_key,
                key=profile_result["key"],
                profile_json=profile_result,
            )
            if response["code"] // 100 != 2:
                adapter.error(
                    "Failed to edit profile after parsing with"
                    f" reference={profile['reference']} response={response}"
                )
                failed.append(profile)

    return failed


def update(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    source_response = hrflow_client.source.get(key=parameters.source_key)

    for profile in items:
        current_profile = hrflow_client.profile.storing.get(
            source_key=parameters.source_key,
            reference=profile["reference"],
        ).get("data")

        if not current_profile:
            adapter.warning(
                f"Profile with reference={profile['reference']} not found in source."
                " Failing for profile..."
            )
            failed.append(profile)
            continue

        edit = (
            {field: profile.get(field) for field in parameters.only_edit_fields}
            if parameters.only_edit_fields
            else profile
        )
        profile_to_edit = {**current_profile, **edit}
        profile_to_edit.pop(
            "resume", None
        )  # remove resume file from profile to edit because it's not JSON serializable

        if profile.get("resume") and profile["resume"].get("raw") is not None:
            if not current_profile.get("attachments"):
                parsing_response = hrflow_client.profile.parsing.add_file(
                    source_key=parameters.source_key,
                    profile_file=profile["resume"]["raw"],
                    profile_content_type=profile["resume"].get("content_type"),
                    profile_file_name=profile["resume"].get("file_name"),
                    reference=profile["reference"],
                    tags=profile.get("tags", []),
                    metadatas=profile.get("metadatas", {}),
                    created_at=profile.get("created_at"),
                )

                if parsing_response["code"] not in [201, 202]:
                    adapter.error(
                        "Failed to parse profile with"
                        f" reference={profile['reference']} response={parsing_response}"
                    )
                    failed.append(profile)
                    continue

                if source_response["data"].get("sync_parsing") is True:
                    parsing_result = parsing_response["data"]["profile"]
                    profile_to_edit = hydrate_profile(parsing_result, profile_to_edit)

        response = hrflow_client.profile.storing.edit(
            source_key=parameters.source_key,
            key=current_profile["key"],
            profile_json=profile_to_edit,
        )

        if response["code"] != 200:
            adapter.error(
                "Failed to edit profile with"
                f" reference={profile_to_edit['reference']}"
                f" key={profile_to_edit['key']} response={response}"
            )
            failed.append(profile)

    return failed


def archive(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ArchiveCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )

    for profile in items:
        profile_reference = profile.get("reference")
        if not profile_reference:
            adapter.error("can't archive profile without reference")
            failed.append(profile)
            continue
        response = hrflow_client.profile.storing.archive(
            source_key=parameters.source_key, reference=profile_reference
        )
        if response["code"] >= 400:
            if "Unable to find object: profile" in response["message"]:
                adapter.error(
                    "Failed to archive profile with reference={} source_key={}"
                    " response={}".format(
                        profile["reference"],
                        parameters.source_key,
                        response,
                    )
                )
                continue
            adapter.error(
                "Failed to archive profile with reference={} source_key={}"
                " response={}".format(
                    profile_reference, parameters.source_key, response
                )
            )
            failed.append(profile)
    return failed


ProfilesAisle = Aisle(
    name=Entity.profile,
    read=ReadOperation(
        function=merge(create=read, update=read, archive=read),
        criterias=Criterias(
            create=ReadAllModesCriterias,
            update=ReadAllModesCriterias,
            archive=ReadAllModesCriterias,
        ),
    ),
    write=WriteOperation(
        function=merge(create=create, update=update, archive=archive),
        criterias=Criterias(
            create=CreateCriterias, update=UpdateCriterias, archive=ArchiveCriterias
        ),
    ),
    schema=HrFlowProfile,
)



import base64
import typing as t
from logging import LoggerAdapter

import requests
from pydantic import Field

from hrflow_connectors.core import (
    ActionEndpoints,
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)
from hrflow_connectors.v1.connectors.greenhouse.schemas import (
    GreenhouseJobModel,
    GreenhouseProfileModel,
)

GET_JOB_ENDPOINT = ActionEndpoints(
    name="Get job",
    description=(
        "Endpoint to get the content of a job with a given id. The request method is"
        " `GET`"
    ),
    url="https://developers.greenhouse.io/harvest.html?shell#get-retrieve-job",
)
POST_CANDIDATE_ENDPOINT = ActionEndpoints(
    name="Post Candidate",
    description=(
        "Endpoint to create a new candidate and assign to a talent pool, the request"
        " method is `POST`"
    ),
    url="https://developers.greenhouse.io/job-board.html#jobs",
)


class WriteProfilesParameters(ParametersModel):
    auth: str = Field(..., description="XAPIKeyAuth", field_type=FieldType.Auth)
    on_behalf_of: str = Field(
        ...,
        description=(
            "The ID of the user sending the profile, or the person he is sending the"
            " profile on behalf of"
        ),
        field_type=FieldType.QueryParam,
    )


class ReadProfilesParameters(ParametersModel):
    auth: str = Field(..., description="XAPIKeyAuth", field_type=FieldType.Auth)
    created_after: str = Field(
        None,
        description=(
            "Return only candidates that were created at or after this timestamp."
            " Timestamp must be in in ISO-8601 format."
        ),
        field_type=FieldType.QueryParam,
    )
    updated_after: str = Field(
        None,
        description=(
            "Return only candidates that were updated at or after this timestamp."
            " Timestamp must be in in ISO-8601 format."
        ),
        field_type=FieldType.QueryParam,
    )
    job_id: str = Field(
        None,
        description=(
            "If supplied, only return candidates that have applied to this job. Will"
            " return both when a candidate has applied to a job and when they’re a"
            " prospect for a job."
        ),
        field_type=FieldType.QueryParam,
    )
    email: str = Field(
        None,
        description=(
            "If supplied, only return candidates who have a matching e-mail address. If"
            " supplied with job_id, only return a candidate with a matching e-mail with"
            " an application on the job. If email and candidate_ids are included,"
            " candidate_ids will be ignored."
        ),
        field_type=FieldType.QueryParam,
    )
    candidate_ids: str = Field(
        None,
        description=(
            "If supplied, only return candidates with matching ids. If supplied with"
            " job_id, only return a candidate with a matching id with an application on"
            " the job. If email and candidate_ids are included, candidate_ids will be"
            " ignored."
        ),
        field_type=FieldType.QueryParam,
    )


class ReadJobsParameters(ParametersModel):
    board_token: str = Field(
        ..., description="Board_token", field_type=FieldType.QueryParam
    )


def read_profiles(
    adapter: LoggerAdapter,
    parameters: ReadProfilesParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    parameters.auth = parameters.auth + ":"
    authorization = base64.b64encode(parameters.auth.encode("ascii"))
    params = parameters.dict()
    del params["auth"]
    response = requests.get(
        "https://harvest.greenhouse.io/v1/candidates",
        headers={
            "Authorization": b"Basic " + authorization,
        },
        params=params,
    )
    if response.status_code // 100 != 2:
        adapter.error(
            "Failed to pull profiles from Greenhouse params={}"
            " status_code={} response={}".format(
                parameters, response.status_code, response.text
            )
        )
        raise Exception("Failed to pull profiles from Greenhouse")
    response = response.json()
    profiles = response["candidates"]
    adapter.info("Pulling {} profiles".format(len(profiles)))
    for profile in profiles:
        yield profile


def read(
    adapter: LoggerAdapter,
    parameters: ReadJobsParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    while True:
        response = requests.get(
            "https://boards-api.greenhouse.io/v1/boards/{}/jobs/?content=true".format(
                parameters.board_token
            ),
            headers={},
            params=None,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Greenhouse params={}"
                " status_code={} response={}".format(
                    parameters, response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Greenhouse")
        response = response.json()
        jobs = response["jobs"]
        if len(jobs) == 0:
            break

        adapter.info("Pulling {} jobs".format(len(jobs)))
        for job in jobs:
            yield job
        break


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    adapter.info("Pushing {} profiles".format(len(profiles)))
    failed_profiles = []
    for profile in profiles:
        parameters.auth = parameters.auth + ":"
        authorization = base64.b64encode(parameters.auth.encode("ascii"))
        response = requests.post(
            "https://harvest.greenhouse.io/v1/candidates",
            headers={
                "On-Behalf-Of": parameters.on_behalf_of,
                "Authorization": b"Basic " + authorization,
            },
            json=profile,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to push profile to Greenhouse status_code={} response={}"
                .format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


GreenhouseJobWarehouse = Warehouse(
    name="Greenhouse Jobs",
    data_schema=GreenhouseJobModel,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read,
        endpoints=[GET_JOB_ENDPOINT],
    ),
)

GreenhouseProfileWarehouse = Warehouse(
    name="Greenhouse Profiles",
    data_schema=GreenhouseProfileModel,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters,
        function=write,
        endpoints=[POST_CANDIDATE_ENDPOINT],
    ),
    read=WarehouseReadAction(
        parameters=ReadProfilesParameters,
        function=read_profiles,
    ),
)



import typing as t
from logging import LoggerAdapter

from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.waalaxy.schemas import WaalaxyProfile
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import Aisle, Criterias, ReadOperation, merge


class AuthParameters(Struct):
    pass


class ReadProfilesParameters(Struct):
    profile: Annotated[
        t.Dict,
        Meta(
            description="Profile object recieved from the Webhook",
        ),
    ]


def read(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    return [parameters.profile]


ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=WaalaxyProfile,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(create=read, update=read, archive=read),
    ),
)



from typing import Any, Optional

from msgspec import Meta, Struct
from typing_extensions import Annotated


class BullhornAddress(Struct):
    address1: Annotated[Optional[str], Meta(description="Adress of the profile")] = None
    city: Annotated[Optional[str], Meta(description="City of the profile")] = None
    state: Annotated[Optional[str], Meta(description="Country code of the profile")] = (
        None
    )
    zip: Annotated[Optional[str], Meta(description="Postal code of the profile")] = None


class BullhornProfile(Struct):
    id: Annotated[
        Optional[str], Meta(description="Unique identifier for this entity")
    ] = None
    address: Annotated[
        Optional[BullhornAddress], Meta(description="Candidate address")
    ] = None
    certifications: Annotated[Any, Meta(description="Candidate’s certifications")] = (
        None
    )
    name: Annotated[
        Optional[str],
        Meta(
            description=(
                "Candidate’s full name. If setting firstname or lastname, you must also"
                " set this field; it does not populate automatically"
            ),
        ),
    ] = None
    firstName: Annotated[Optional[str], Meta(description="Candidate’s first name")] = (
        None
    )
    lastName: Annotated[Optional[str], Meta(description="Name of the file")] = None
    email: Annotated[
        Optional[Optional[str]], Meta(description="Candidate’s email address")
    ] = None
    mobile: Annotated[
        Optional[Optional[str]],
        Meta(description="Candidate’s mobile (cell) telephone number"),
    ] = None
    dateOfBirth: Annotated[
        Optional[int], Meta(description="Candidate’s date of birth")
    ] = None
    experience: Annotated[
        Optional[int],
        Meta(description="Number of years of experience that the Candidate has"),
    ] = None
    skillSet: Annotated[
        Optional[str], Meta(description="Text description of Candidate’s skills")
    ] = None


class BullhornAttachmentEnrichment(Struct, kw_only=True):
    externalID: Annotated[
        Optional[str], Meta(description="External identifier for the file")
    ] = None
    fileContent: Annotated[
        Optional[str],
        Meta(description="Base64-encoded Optional[str]ing of the file content"),
    ] = None
    fileExtension: Annotated[
        Optional[Optional[str]],
        Meta(description="Extension of the file. For example, .doc or .jpg"),
    ] = None
    fileType: Annotated[
        Optional[str], Meta(description="Always use the value “SAMPLE”")
    ] = None
    name: Annotated[
        Optional[str],
        Meta(
            description=(
                "File name. If a file extension is included as part of the name and the"
                " fileExtension field is not set, the file extension in the name is"
                " used."
            ),
        ),
    ] = None
    contentType: Annotated[
        Optional[str], Meta(description="Type/subtype of the file content.type")
    ] = None
    description: Annotated[
        Optional[str], Meta(description="Unique identifier for this entity")
    ] = None
    type: Optional[str]


class BullhornCandidate(Struct):
    id: Annotated[
        Optional[Optional[int]], Meta(description="Unique identifier for this entity")
    ] = None


class BullhornExperienceEnrichment(Struct, kw_only=True):
    id: Annotated[
        Optional[str], Meta(description="Unique identifier for this entity")
    ] = None
    candidate: Annotated[
        BullhornCandidate,
        Meta(description="Candidate for whom this person is a reference"),
    ]
    companyName: Annotated[
        Optional[Optional[str]],
        Meta(
            description=(
                "Name of the company where reference works, if it does not "
                "have a ClientCorporation record in Bullhorn"
            ),
        ),
    ] = None
    title: Annotated[
        Optional[Optional[str]],
        Meta(description="Candidate’s job title in this position"),
    ] = None
    comments: Annotated[
        Optional[Optional[str]],
        Meta(description="Free-text comments on CandidateWorkHistory"),
    ] = None
    startDate: Annotated[
        Optional[Optional[int]],
        Meta(description="Date on which Candidate began working at this position"),
    ] = None
    endDate: Annotated[
        Optional[Optional[int]],
        Meta(description="Date on which job ended, if applicable"),
    ] = None


class BullhornEducationEnrichment(Struct, kw_only=True):
    id: Annotated[
        Optional[str], Meta(description="Unique identifier for this entity")
    ] = None
    candidate: Annotated[
        BullhornCandidate,
        Meta(description="Candidate for whom this person is a reference"),
    ]
    school: Annotated[
        Optional[Optional[str]],
        Meta(
            description=(
                "Name of the educational institute where this education took place"
            )
        ),
    ] = None
    degree: Annotated[
        Optional[Optional[str]],
        Meta(
            description=(
                "Indicates what educational degree the Candidate received; for "
                "example, B.A., M.A., Ph.D., and so forth"
            ),
        ),
    ] = None
    comments: Annotated[
        Optional[Optional[str]], Meta(description="Free-text comments on this record")
    ] = None
    city: Annotated[
        Optional[Optional[str]],
        Meta(description="Name of the city where the education took place"),
    ] = None
    startDate: Annotated[
        Optional[Optional[int]], Meta(description="Date when Candidate began study")
    ] = None
    endDate: Annotated[
        Optional[Optional[int]],
        Meta(description="Date when Candidate finished this education"),
    ] = None


class BullhornJob(Struct):
    id: Annotated[
        Optional[Optional[int]], Meta(description="Unique identifier for this entity")
    ] = None



import json
import os
import typing as t
from abc import abstractmethod
from collections import Counter
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum
from importlib import import_module
from pathlib import Path

import yaml
from msgspec import Struct, ValidationError, convert, field

from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.connector import Connector
from hrflow_connectors.v2.core.run import Event, Reason, Status
from hrflow_connectors.v2.core.utils import CONNECTORS_DIRECTORY
from hrflow_connectors.v2.core.warehouse import Warehouse

PROJECT_DIRECTORY = Path(__file__).parent.parent.parent.parent.parent

SECRETS_PREFIX = "$__"
ENVIRON_SECRETS_PREFIX = "HRFLOW_CONNECTORS_V2_{connector_name}_"
Secrets = ContextVar("secrets")
ConnectorName = ContextVar("connector_name")
ActionNames = ContextVar("action_names")
WarehouseNames = ContextVar("warehouse_names")


class NoTestConfigException(Exception):
    pass


class InvalidJSONException(Exception):
    pass


class InvalidYAMLException(Exception):
    pass


class InvalidTestConfigException(Exception):
    pass


@contextmanager
def secrets(connector_name: str, connector_subtype: str, connectors_directory: Path):
    secrets_prefix = ENVIRON_SECRETS_PREFIX.format(
        connector_name=connector_name.upper()
    )

    global_secrets_file = connectors_directory / "secrets.json"
    if global_secrets_file.exists():
        try:
            global_secrets = json.loads(global_secrets_file.read_text())
            global_secrets = {
                key.replace(secrets_prefix, "", 1): value
                for key, value in global_secrets.items()
                if key.startswith(secrets_prefix)
            }
        except json.JSONDecodeError as e:
            raise InvalidJSONException(
                "Failed to JSON decode global secrets file for connector {} with"
                " error {}".format(connector_name, e)
            )
    else:
        global_secrets = dict()

    connector_secrets_file = connectors_directory / connector_subtype / "secrets.json"
    if connector_secrets_file.exists():
        try:
            connector_secrets = json.loads(connector_secrets_file.read_text())
        except json.JSONDecodeError as e:
            raise InvalidJSONException(
                "Failed to JSON decode secrets file for connector {} with error {}"
                .format(connector_name, e)
            )
    else:
        connector_secrets = dict()

    environnment_secrets = {
        key.replace(secrets_prefix, "", 1): value
        for key, value in os.environ.items()
        if key.startswith(secrets_prefix)
    }
    token = Secrets.set({**global_secrets, **connector_secrets, **environnment_secrets})
    yield
    Secrets.reset(token)


@contextmanager
def actions(connector: Connector):
    connector_name_token = ConnectorName.set(connector.name)
    actions_token = ActionNames.set(
        [flow.name(connector.subtype) for flow in connector.flows]
    )
    yield
    ConnectorName.reset(connector_name_token)
    ActionNames.reset(actions_token)


@contextmanager
def warehouses(connector_subtype: str, connectors_directory: Path):
    if connectors_directory is CONNECTORS_DIRECTORY:  # pragma: no cover
        warehouse_module = import_module(
            "hrflow_connectors.v2.connectors.{}.warehouse".format(connector_subtype)
        )
    else:
        import_from = connectors_directory.relative_to(PROJECT_DIRECTORY)
        warehouse_module = import_module(
            "{}.{}.warehouse".format(
                str(import_from).replace("/", "."), connector_subtype
            )
        )
    warehouse_names = []
    for attribute, value in warehouse_module.__dict__.items():
        if isinstance(value, Warehouse):
            warehouse_names.append(attribute)
    token = WarehouseNames.set(warehouse_names)
    yield
    WarehouseNames.reset(token)


class Validated:
    @classmethod
    @abstractmethod
    def validate(cls, value: t.Any) -> t.Any:
        raise NotImplementedError()  # pragma: nocover


# This was added to avoid failure during
# msgspec convert. What happens is that
# msgspec is expecting instances of ParameterValue
# but ParameterValue is more a validation wrapper
# with no type as if depends on what might be present
# in secrets
class NeverFailIsInstanceMeta(type):
    def __instancecheck__(cls, inst):
        return True


class ParameterValue(Validated, metaclass=NeverFailIsInstanceMeta):
    @classmethod
    def validate(cls, value: t.Any):
        if isinstance(value, str) and value.startswith(SECRETS_PREFIX):
            key = value.replace(SECRETS_PREFIX, "", 1)
            if key not in Secrets.get(dict()):
                raise TypeError(
                    "'{}' not found in secrets for connector {}".format(
                        key, ConnectorName.get("")
                    )
                )
            return Secrets.get()[key]
        return value


class ActionName(Validated):
    @classmethod
    def validate(cls, value: t.Any):  # pragma: nocover
        action_names = ActionNames.get([])
        if value not in action_names:
            raise TypeError(
                "No action '{}' found for connector {}. Should be one of {}".format(
                    value, ConnectorName.get(""), action_names
                )
            )
        return value


class WarehouseName(Validated):
    @classmethod
    def validate(cls, value: t.Any):  # pragma: nocover
        warehoue_names = WarehouseNames.get([])
        if value not in warehoue_names:
            raise TypeError(
                "Warehouse '{}' not found for connector {}. Should be one of {}".format(
                    value, ConnectorName.get(""), warehoue_names
                )
            )
        return value


class ReadTest(Struct):
    mode: Mode
    entity: Entity
    auth_parameters: dict[str, ParameterValue]
    id: t.Optional[str] = None
    parameters: dict[str, ParameterValue] = field(default_factory=dict)
    incremental: bool = False
    incremental_token: t.Optional[str] = None
    expected_number_of_items: t.Optional[int] = None


class WarehouseTests(Struct):
    read: list[ReadTest] = field(default_factory=list)


class ActionTest(Struct):
    connector_auth: dict[str, ParameterValue]
    hrflow_auth: dict[str, ParameterValue]
    id: t.Optional[str] = None
    pull_parameters: dict[str, ParameterValue] = field(default_factory=dict)
    push_parameters: dict[str, ParameterValue] = field(default_factory=dict)
    status: t.Optional[Status] = None
    reason: t.Optional[Reason] = None
    events: t.Optional[t.Counter[Event]] = None


class ConnectorTestConfig(Struct):
    if t.TYPE_CHECKING:
        warehouse: dict[str, WarehouseTests] = field(
            default_factory=dict
        )  # pragma: nocover
        actions: dict[str, list[ActionTest]] = field(
            default_factory=dict
        )  # pragma: nocover
    else:
        warehouse: dict[WarehouseName, WarehouseTests] = field(default_factory=dict)
        actions: dict[ActionName, list[ActionTest]] = field(default_factory=dict)


def dec_hook(type: type, obj: t.Any) -> t.Any:
    if (
        t.get_origin(type) is Counter
        and len(t.get_args(type)) == 1
        and issubclass((EnumModel := t.get_args(type)[0]), Enum)
    ):
        return Counter({EnumModel(key): value for key, value in obj.items()})

    if issubclass(type, Validated):
        return type.validate(obj)

    else:
        raise NotImplementedError(
            f"Objects of type {type} are not supported"
        )  # pragma: nocover


def collect_connector_tests(
    connector: Connector, connectors_directory: Path = CONNECTORS_DIRECTORY
):
    connector_name = connector.name
    connector_subtype = connector.subtype
    test_config_file = connectors_directory / connector_subtype / "test-config.yaml"
    if test_config_file.exists() is False:
        raise NoTestConfigException(
            "No test configuration found for connector {} at {}".format(
                connector_name, test_config_file
            )
        )
    try:
        test_config = yaml.safe_load(test_config_file.read_bytes())
    except yaml.YAMLError as e:
        raise InvalidYAMLException(
            "Failed to parse yaml test config for connector {} with error {} ".format(
                connector_name, str(e)
            )
        )

    try:
        with warehouses(
            connector_subtype=connector_subtype,
            connectors_directory=connectors_directory,
        ):
            with actions(
                connector=connector,
            ):
                with secrets(
                    connector_name=connector_name,
                    connector_subtype=connector_subtype,
                    connectors_directory=connectors_directory,
                ):
                    test_suite = convert(
                        test_config, ConnectorTestConfig, dec_hook=dec_hook
                    )
    except ValidationError as e:
        raise InvalidTestConfigException(e.args)

    return test_suite



import typing as t

from hrflow_connectors.core import (
    ActionName,
    ActionType,
    BaseActionParameters,
    Connector,
    ConnectorAction,
    ConnectorType,
    WorkflowType,
)
from hrflow_connectors.v1.connectors.hrflow.warehouse import (
    HrFlowProfileParsingWarehouse,
)
from hrflow_connectors.v1.connectors.meteojob.warehouse import MeteojobProfilesWarehouse


def rename_profile_fields(meteojob_profile: t.Dict) -> t.Dict:
    return {
        "job-number": meteojob_profile.get("jobkey", "")[:10] or None,
        "first_name": meteojob_profile.get("firstName"),
        "last_name": meteojob_profile.get("lastName"),
        "phone": meteojob_profile.get("phone"),
        "email": meteojob_profile.get("email"),
        "coverText": meteojob_profile.get("coverText"),
        "profile-country": meteojob_profile.get("profilecountry"),
        "profile-regions": meteojob_profile.get("profileregions"),
        "profile-domains": meteojob_profile.get("profiledomains"),
        "job-lien_annonce_site_carriere": meteojob_profile.get(
            "joblien_annonce_site_carriere"
        ),
        "statistic-source": meteojob_profile.get("statisticsource"),
        "statistic-jbsource": meteojob_profile.get("statisticjbsource"),
    }


def add_tags(profile_tags: t.Dict) -> t.List[t.Dict]:
    return [dict(name=key, value=value) for key, value in profile_tags.items() if value]


def format_meteojob_profile(meteojob_profile: t.List) -> t.Dict:
    profile_tags = rename_profile_fields(meteojob_profile)
    tags = add_tags(profile_tags)
    resume_dict = dict(
        raw=meteojob_profile["cv"],
        content_type=meteojob_profile["content_type"],
    )
    return dict(
        reference=None,
        resume=resume_dict,
        tags=tags,
        metadatas=[],
        created_at=None,
    )


def event_parser(event: t.Dict) -> t.Dict:
    return dict(profile=event)


DESCRIPTION = (
    "METEOJOB : LEADER EUROPÉEN DU MATCHING, BIG DATA ET DES ENTRETIENS VIDÉO DANS"
    " LES RH "
)
Meteojob = Connector(
    name="Meteojob",
    type=ConnectorType.JobBoard,
    subtype="meteojob",
    description=DESCRIPTION,
    url="https://www.meteojob.com/",
    actions=[
        ConnectorAction(
            name=ActionName.catch_profile,
            trigger_type=WorkflowType.catch,
            description="Imports candidates, in synchronization with Meteojob",
            parameters=BaseActionParameters.with_defaults(
                "TriggerViewActionParameters",
                format=format_meteojob_profile,
                event_parser=event_parser,
            ),
            origin=MeteojobProfilesWarehouse,
            target=HrFlowProfileParsingWarehouse,
            action_type=ActionType.inbound,
        )
    ],
)



import html
import re
import typing as t
from logging import LoggerAdapter

from hrflow import Hrflow
from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.hrflow.aisles.common import AuthParameters
from hrflow_connectors.v2.core.hrflow.schemas import HrFlowJob
from hrflow_connectors.v2.core.warehouse import Aisle, Criterias, WriteOperation, merge

LIST_JOBS_LIMIT = 30

LABEL_TO_JOB_FIELD = dict(
    language="languages",
    task="tasks",
    certification="certifications",
    course="courses",
    interest="interests",
)
SKILL_LABEL_TO_TYPE = dict(Skill=None, skill_hard="hard", skill_soft="soft")


class JobParsingException(Exception):
    def __init__(self, *args, client_response: dict):
        self.client_response = client_response


class CreateCriterias(Struct):
    board_key: Annotated[
        str,
        Meta(
            description="HrFlow.ai board key",
        ),
    ]
    enrich_with_parsing: Annotated[
        bool,
        Meta(
            description="When enabled jobs are enriched with HrFlow.ai parsing",
        ),
    ] = False


class UpdateCriterias(Struct):
    board_key: Annotated[str, Meta(description="HrFlow.ai board key")]


class ArchiveCriterias(Struct):
    board_key: Annotated[str, Meta(description="HrFlow.ai board key")]


def remove_html_tags(text: str) -> str:
    return re.sub("<[^<]+?>", "", text)


def enrich_job_with_parsing(hrflow_client: Hrflow, job: dict) -> None:
    concatenate = []
    summary = job.get("summary")
    if summary:
        concatenate.append(summary)

    for section in job.get("sections") or []:
        content = []
        title = section.get("title")
        if title:
            content.append(title)
        description = section.get("description")
        if description:
            content.append(description)
        if content:
            concatenate.append("\n".join(content))

    concatenated = "\n\n".join(concatenate)
    cleaned = html.unescape(remove_html_tags(concatenated)).strip()
    if cleaned == "":
        return

    response = hrflow_client.text.parsing.post(texts=[cleaned])
    if response["code"] >= 400:
        raise JobParsingException("Failed to parse job", client_response=response)

    entities, parsed_text = response["data"][0]["entities"], response["data"][0]["text"]
    for attribute in [
        "skills",
        "languages",
        "tasks",
        "certifications",
        "courses",
        "interests",
    ]:
        if job.get(attribute) is None:
            job[attribute] = []

    for entitiy in entities:
        label = entitiy["label"]
        entity_text = parsed_text[entitiy["start"] : entitiy["end"]]

        if label in LABEL_TO_JOB_FIELD:
            job_field = LABEL_TO_JOB_FIELD[label]
            if next(
                (
                    element
                    for element in job[job_field]
                    if element["name"] == entity_text
                ),
                False,
            ):
                continue
            job[job_field].append(dict(name=entity_text, value=None))
        elif label in SKILL_LABEL_TO_TYPE:
            skill_type = SKILL_LABEL_TO_TYPE[label]
            if next(
                (
                    skill
                    for skill in job["skills"]
                    if skill["name"] == entity_text and skill["type"] == skill_type
                ),
                False,
            ):
                continue
            job["skills"].append(dict(name=entity_text, type=skill_type, value=None))

    return


def create(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: CreateCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed_jobs = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    for job in items:
        if parameters.enrich_with_parsing:
            adapter.info("Starting parsing for job without reference")
            try:
                enrich_job_with_parsing(hrflow_client, job)
                adapter.info("Parsing finished")
            except JobParsingException as e:
                adapter.error(
                    "Failed to parse job response={}".format(e.client_response)
                )
                failed_jobs.append(job)
                continue
        response = hrflow_client.job.storing.add_json(
            board_key=parameters.board_key, job_json=job
        )
        if response["code"] >= 400:
            adapter.error(
                "Failed to index job with reference={} board_key={} response={}".format(
                    job["reference"], parameters.board_key, response
                )
            )
            failed_jobs.append(job)
    return failed_jobs


def update(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed_jobs = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    for job in items:
        job_reference = job.get("reference")
        job_key = job.get("key")
        if job_reference is None and job_key is None:
            adapter.error("can't update job without reference or key")
            failed_jobs.append(job)

        response = hrflow_client.job.storing.edit(
            board_key=parameters.board_key, job_json=job
        )
        if response["code"] >= 400:
            if "Unable to find object: job" in response["message"]:
                adapter.error(
                    "Failed to update job with reference={} board_key={} response={}"
                    .format(job["reference"], parameters.board_key, response)
                )
                continue
            adapter.error(
                "Failed to update job with reference={} board_key={} response={}"
                .format(job["reference"], parameters.board_key, response)
            )
            failed_jobs.append(job)

    return failed_jobs


def archive(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed_jobs = []
    hrflow_client = Hrflow(
        api_secret=auth_parameters.api_secret, api_user=auth_parameters.api_user
    )
    for job in items:
        job_reference = job.get("reference")

        if not job_reference:
            adapter.error("can't archive job without reference")
            failed_jobs.append(job)
            continue
        response = hrflow_client.job.storing.archive(
            board_key=parameters.board_key, reference=job_reference
        )
        if response["code"] >= 400:
            if "Unable to find object: job" in response["message"]:
                adapter.error(
                    "Failed to archive job with reference={} board_key={} response={}"
                    .format(job["reference"], parameters.board_key, response)
                )
                continue
            adapter.error(
                "Failed to archive job with reference={} board_key={} response={}"
                .format(job_reference, parameters.board_key, response)
            )
            failed_jobs.append(job)

    return failed_jobs


JobsAisle = Aisle(
    name=Entity.job,
    write=WriteOperation(
        function=merge(create=create, update=update, archive=archive),
        criterias=Criterias(
            create=CreateCriterias, update=UpdateCriterias, archive=ArchiveCriterias
        ),
    ),
    schema=HrFlowJob,
)



import typing as t

from hrflow_connectors.v2.connectors.adzuna.warehouse import AdzunaWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_tags(adzuna_job: t.Dict) -> t.List[t.Dict]:
    t = lambda name, value: dict(name=name, value=value)
    tags = [
        t("contract_type", adzuna_job.get("contract_type")),
        t("contract_time", adzuna_job.get("contract_time")),
        t("salary_min", adzuna_job.get("salary_min")),
        t("salary_max", adzuna_job.get("salary_max")),
        t(
            "salary_is_predicted",
            adzuna_job.get("salary_is_predicted", 0) == 1,
        ),
        t("category", adzuna_job.get("category", {}).get("label")),
        t("company", adzuna_job.get("company", {}).get("display_name")),
    ]
    return tags


def format_job(
    adzuna_job: t.Dict,
) -> t.Dict:
    job = dict(
        name=adzuna_job.get("title"),
        reference=str(adzuna_job.get("id")),
        created_at=adzuna_job.get("created"),
        location=dict(
            lat=adzuna_job.get("latitude"),
            lng=adzuna_job.get("longitude"),
            text=adzuna_job.get("location", {}).get("display_name", ""),
        ),
        url=adzuna_job.get("redirect_url"),
        summary=adzuna_job.get("description"),
        sections=(
            [
                dict(
                    name="full_description",
                    title="Full Description",
                    description=adzuna_job.get("full_description"),
                )
            ]
            if adzuna_job.get("full_description")
            else []
        ),
        tags=get_tags(adzuna_job),
    )
    return job


def format_archive(adzuna_job: t.Dict) -> t.Dict:
    return dict(reference=str(adzuna_job.get("id")))


DESCRIPTION = (
    "Find Every Job, Everywhere with Adzuna\n\nAdzuna is a smarter, more transparent"
    " job search engine that helps you dodge the thousands of irrelevant jobs so you"
    " can zero in on the right role faster."
)
Adzuna = Connector(
    name="Adzuna",
    type=ConnectorType.JobBoard,
    subtype="adzuna",
    description=DESCRIPTION,
    url="https://www.adzuna.fr/",
    warehouse=AdzunaWarehouse,
    flows=(
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.archive, Entity.job, Direction.inbound, format=format_archive),
    ),
)



import typing as t
from logging import LoggerAdapter

from msgspec import Meta, Struct
from smbprotocol.open import (
    CreateDisposition,
    CreateOptions,
    FilePipePrinterAccessMask,
    ImpersonationLevel,
    Open,
    ShareAccess,
)
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.admen.schemas import AdmenMission, AdmenProfile
from hrflow_connectors.v2.connectors.admen.utils.mysql_utils import (
    connect_to_database,
    insert_object,
    query_database,
    update_object,
)
from hrflow_connectors.v2.connectors.admen.utils.smb_utils import (
    establish_smb_connection,
)
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)


class AuthParameters(Struct):
    db_host: Annotated[
        str,
        Meta(
            description="The hostname of the database server",
        ),
    ]
    db_port: Annotated[
        int,
        Meta(
            description="The port of the database server",
        ),
    ]
    db_name: Annotated[
        str,
        Meta(
            description="The name of the database",
        ),
    ]
    db_user: Annotated[
        str,
        Meta(
            description="The username to connect to the database",
        ),
    ]
    db_password: Annotated[
        str,
        Meta(
            description="The password to connect to the database",
        ),
    ]


class ReadJobsParameters(Struct):
    limit: Annotated[
        t.Optional[int],
        Meta(
            description="The maximum number of items to fetch",
        ),
    ] = None


class ReadProfilesParameters(Struct):
    share_server: Annotated[
        t.Optional[str],
        Meta(
            description="The hostname of the network share server",
        ),
    ] = None
    share_name: Annotated[
        t.Optional[str],
        Meta(
            description="The name of the network share",
        ),
    ] = None
    share_username: Annotated[
        t.Optional[str],
        Meta(
            description="The username to connect to the network share",
        ),
    ] = None
    share_password: Annotated[
        t.Optional[str],
        Meta(
            description="The password to connect to the network share",
        ),
    ] = None
    limit: Annotated[
        t.Optional[int],
        Meta(
            description="The maximum number of items to fetch",
        ),
    ] = None


class WriteProfilesParameters(Struct):
    pass


def generic_job_read(
    mode: Mode,
):
    def read_jobs(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadJobsParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[t.Dict]:
        jobs_table = "MISSIONS"
        query_conditions = {
            Mode.create: "WHERE DATE_MAJ = DATE_SAISIE",
            Mode.update: "WHERE DATE_MAJ > DATE_SAISIE",
            Mode.archive: "WHERE SOFT_DELETED = 1",
        }
        query = f"SELECT * FROM {jobs_table} {query_conditions[mode]}"
        if parameters.limit:
            query += f" LIMIT {parameters.limit}"

        adapter.info(f"Executing query: {query}")

        # Connect to the database
        connection = connect_to_database(
            auth_parameters.db_host,
            auth_parameters.db_port,
            auth_parameters.db_user,
            auth_parameters.db_password,
            auth_parameters.db_name,
        )
        if connection:
            jobs = query_database(connection, query)
            job_count = len(jobs) if jobs else 0
            adapter.info(f"Found {job_count} jobs in the database")
            if jobs:
                yield from jobs
            connection.close()

    return read_jobs


def fetch_and_save_document(adapter, tree, file_path):
    try:
        # Open the document and read its content
        document = Open(tree, file_path)
        document.create(
            impersonation_level=ImpersonationLevel.Impersonation,
            desired_access=FilePipePrinterAccessMask.FILE_READ_DATA,
            file_attributes=0,
            share_access=ShareAccess.FILE_SHARE_READ,
            create_disposition=CreateDisposition.FILE_OPEN,
            create_options=CreateOptions.FILE_NON_DIRECTORY_FILE,
        )
        file_size = document.end_of_file
        read_offset = 0
        read_length = file_size
        file_data = document.read(read_offset, read_length)
        document.close()

        # Return the raw content and content type
        return {"raw": file_data, "content_type": "pdf"}

    except Exception as e:
        adapter.error(f"An error occurred while fetching the document: {e}")
        return None


def generic_profile_read(
    mode: Mode,
):
    """Factory function to read profiles based on the specified mode."""

    def read_profiles(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: ReadProfilesParameters,
        incremental: bool,
        incremental_token: t.Optional[str],
        max_retries: int = 5,
        retry_delay: int = 5,
    ) -> t.Iterable[t.Dict]:
        profiles_table = "PERSONNES"
        experiences_table = "EXPERIENCES_PROFESSIONNELLES"
        documents_table = "DOCUMENTS"

        query_conditions = {
            Mode.create: "WHERE DATE_MAJ = DATE_CREATION",
            Mode.update: "WHERE DATE_MAJ > DATE_CREATION",
            Mode.archive: "WHERE SOFT_DELETED = 1",
        }
        query = f"SELECT * FROM {profiles_table} {query_conditions[mode]}"
        if parameters.limit:
            query += f" LIMIT {parameters.limit}"

        adapter.info(f"Executing query: {query}")

        connection = connect_to_database(
            auth_parameters.db_host,
            auth_parameters.db_port,
            auth_parameters.db_user,
            auth_parameters.db_password,
            auth_parameters.db_name,
        )
        if connection:
            profiles = query_database(connection, query)

            # Try to establish an SMB connection
            smb_connected, smb_tree, smb_session, smb_connection = (
                False,
                None,
                None,
                None,
            )
            if (
                parameters.share_server
                and parameters.share_name
                and parameters.share_username
                and parameters.share_password
            ):
                # Establish SMB connection
                smb_connected, smb_tree, smb_session, smb_connection = (
                    establish_smb_connection(
                        adapter,
                        parameters.share_server,
                        parameters.share_username,
                        parameters.share_password,
                        parameters.share_name,
                        max_retries,
                        retry_delay,
                    )
                )

            if profiles:
                for profile in profiles:
                    ID_PERSONNE = profile["ID_PERSONNE"]

                    # Fetch experiences
                    experiences_query = (
                        f"SELECT * FROM {experiences_table} WHERE ID_PERSONNE ="
                        f" {ID_PERSONNE}"
                    )
                    profile["experiences"] = query_database(
                        connection, experiences_query
                    )

                    # Fetch resume
                    documents_query = (
                        f"SELECT * FROM {documents_table} WHERE ID_PERSONNE ="
                        f" '{ID_PERSONNE}' LIMIT 1"
                    )
                    document = query_database(connection, documents_query)
                    if document and smb_connected:
                        profile["resume"] = fetch_and_save_document(
                            adapter,
                            smb_tree,
                            document[0]["PATHNAME"].replace("/", "\\"),
                        )
                    else:
                        profile["resume"] = None

                    yield profile

            # Ensure all connections are closed
            if connection:
                connection.close()

            if smb_connected:
                if smb_tree:
                    smb_tree.disconnect()
                if smb_session:
                    smb_session.disconnect()
                if smb_connection:
                    smb_connection.disconnect()

    return read_profiles


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    # Connect to the database
    connection = connect_to_database(
        auth_parameters.db_host,
        auth_parameters.db_port,
        auth_parameters.db_user,
        auth_parameters.db_password,
        auth_parameters.db_name,
    )
    if connection:
        # Table name
        profiles_table = "PERSONNES"
        experiences_table = "EXPERIENCES_PROFESSIONNELLES"
        # Insert profiles into the database
        for profile in items:
            profile_experiences = profile.pop("experiences", [])
            if not profile.get("scsync_modified"):
                profile["scsync_modified"] = 0
            try:
                inserted_id = insert_object(
                    adapter, connection, profiles_table, profile
                )
                if inserted_id:
                    for experience in profile_experiences:
                        experience["ID_PERSONNE"] = inserted_id
                        insert_object(
                            adapter, connection, experiences_table, experience
                        )
            except Exception as e:
                adapter.error(f"Failed to insert profile into the database: {e}")
                failed_profiles.append(profile)
        connection.close()
    return failed_profiles


def generic_update():
    def update(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: WriteProfilesParameters,
        items: t.Iterable[t.Dict],
    ) -> t.List[t.Dict]:
        failed_items = []
        profiles_table = "PERSONNES"
        experiences_table = "EXPERIENCES_PROFESSIONNELLES"

        connection = connect_to_database(
            auth_parameters.db_host,
            auth_parameters.db_port,
            auth_parameters.db_user,
            auth_parameters.db_password,
            auth_parameters.db_name,
        )
        if connection:
            for item in items:
                ID_PERSONNE = item.pop("ID_PERSONNE")
                experiences = item.pop("experiences", [])
                where_clause = {"ID_PERSONNE": ID_PERSONNE}
                update_result = update_object(
                    adapter, connection, profiles_table, item, where_clause
                )
                if not update_result:
                    failed_items.append(item)
                for experience in experiences:
                    query = f"""
                    SELECT * FROM {experiences_table}
                    WHERE INTITULE_POSTE = %s AND ID_PERSONNE = %s
                    """
                    params = (experience["INTITULE_POSTE"], ID_PERSONNE)
                    existing_experience = query_database(connection, query, params)
                    if existing_experience:
                        where_clause = {
                            "ID_PERSONNE": ID_PERSONNE,
                            "INTITULE_POSTE": experience["INTITULE_POSTE"],
                        }
                        update_result = update_object(
                            adapter,
                            connection,
                            experiences_table,
                            experience,
                            where_clause,
                        )
                        if not update_result:
                            adapter.error(f"Failed to update experience: {experience}")
                    else:
                        experience["ID_PERSONNE"] = ID_PERSONNE
                        insert_object(
                            adapter, connection, experiences_table, experience
                        )

            connection.close()
        return failed_items

    return update


JobsAisle = Aisle(
    name=Entity.job,
    schema=AdmenMission,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=generic_job_read(Mode.create),
            update=generic_job_read(Mode.update),
            archive=generic_job_read(Mode.archive),
        ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=AdmenProfile,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=generic_profile_read(Mode.create),
            update=generic_profile_read(Mode.update),
            archive=generic_profile_read(Mode.archive),
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=WriteProfilesParameters,
            archive=WriteProfilesParameters,
        ),
        function=merge(
            create=write_profiles,
            update=generic_update(),
            archive=generic_update(),
        ),
    ),
)



from msgspec import Meta, Struct
from typing_extensions import Annotated


class AuthParameters(Struct):
    api_secret: Annotated[str, Meta(description="API Key used to access HrFlow.ai API")]
    api_user: Annotated[
        str, Meta(description="User email used to access HrFlow.ai API")
    ]



from __future__ import annotations

import inspect
import re
import typing as t
from dataclasses import dataclass

from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.core.common import Direction
from hrflow_connectors.v2.core.context import MAIN_IMPORT_NAME
from hrflow_connectors.v2.core.hrflow import HrFlowWarehouse
from hrflow_connectors.v2.core.msgspec_pydantic_compat import fields, template_fields
from hrflow_connectors.v2.core.run import CallbackT, FormatT, LogicsT
from hrflow_connectors.v2.core.templates import Templates
from hrflow_connectors.v2.core.utils import get_import_name, reindent_function_source

if t.TYPE_CHECKING:
    from hrflow_connectors.v2.core.connector import Connector, Flow  # pragma: nocover


@dataclass(frozen=True)
class WORKFLOW:
    WORKFLOW_ID_SETTINGS_KEY = "__workflow_id"
    INCREMENTAL_SETTINGS_KEY = "__incremental"

    CONNECTOR_AUTH_SETTINGS_PREFIX = "connector_auth_"
    HRFLOW_AUTH_SETTINGS_PREFIX = "hrflow_auth_"
    PULL_PARAMETERS_SETTINGS_PREFIX = "pull_parameters_"
    PUSH_PARAMETERS_SETTINGS_PREFIX = "push_parameters_"

    LOGICS_PLACEHOLDER = "# << logics_placeholder >>"
    FORMAT_PLACEHOLDER = "# << format_placeholder >>"
    CALLBACK_PLACEHOLDER = "# << callback_placeholder >>"
    EVENT_PARSER_PLACEHOLDER = "# << event_parser_placeholder >>"

    ACTIVATE_INCREMENTAL = "enable"
    DEFAULT_EVENT_PARSER_FUNCTION_NAME = "default_event_parser"
    USER_EVENT_PARSER_FUNCTION_NAME = "event_parser"
    LOGICS_FUNCTIONS_NAME = "logics"
    FORMAT_FUNCTION_NAME = "format"
    CALLBACK_FUNCTION_NAME = "callback"


def workflow(
    connector: Connector,
    flow: Flow,
    integration: t.Literal["catch", "pull"],
) -> str:
    connector_aisle = connector.warehouse.get_aisle(flow.entity)
    hrflow_aisle = HrFlowWarehouse.get_aisle(flow.entity)

    # This is only called with a properly validated
    # connector for which below must be true
    assert connector_aisle is not None
    assert hrflow_aisle is not None

    if flow.direction is Direction.inbound:
        origin_parameters = connector_aisle.parameters("read", flow.mode)
        target_parameters = hrflow_aisle.parameters("write", flow.mode)
    else:
        origin_parameters = hrflow_aisle.parameters("read", flow.mode)
        target_parameters = connector_aisle.parameters("write", flow.mode)

    # This is only called with a properly validated
    # connector for which below must be true
    assert origin_parameters is not None
    assert target_parameters is not None

    default_event_parser = ""
    if flow.event_parser is not None:
        default_event_parser = inspect.getsource(flow.event_parser).replace(
            flow.event_parser.__name__, WORKFLOW.DEFAULT_EVENT_PARSER_FUNCTION_NAME
        )
        default_event_parser = reindent_function_source(
            default_event_parser,
            function_name=WORKFLOW.DEFAULT_EVENT_PARSER_FUNCTION_NAME,
        )

    return Templates.get_template("workflow.py.j2").render(
        main_module=MAIN_IMPORT_NAME.get(),
        import_name=get_import_name(connector),
        workflow_id_settings_key=WORKFLOW.WORKFLOW_ID_SETTINGS_KEY,
        incremental_settings_key=WORKFLOW.INCREMENTAL_SETTINGS_KEY,
        activate_incremental_token=WORKFLOW.ACTIVATE_INCREMENTAL,
        connector_auth_settings_prefix=WORKFLOW.CONNECTOR_AUTH_SETTINGS_PREFIX,
        hrflow_auth_settings_prefix=WORKFLOW.HRFLOW_AUTH_SETTINGS_PREFIX,
        pull_parameters_settings_prefix=WORKFLOW.PULL_PARAMETERS_SETTINGS_PREFIX,
        push_parameters_settings_prefix=WORKFLOW.PUSH_PARAMETERS_SETTINGS_PREFIX,
        logics_placeholder=WORKFLOW.LOGICS_PLACEHOLDER,
        logics_functions_name=WORKFLOW.LOGICS_FUNCTIONS_NAME,
        format_placeholder=WORKFLOW.FORMAT_PLACEHOLDER,
        format_function_name=WORKFLOW.FORMAT_FUNCTION_NAME,
        callback_placeholder=WORKFLOW.CALLBACK_PLACEHOLDER,
        callback_function_name=WORKFLOW.CALLBACK_FUNCTION_NAME,
        event_parser_placeholder=WORKFLOW.EVENT_PARSER_PLACEHOLDER,
        default_event_parser=default_event_parser,
        user_event_parser_function_name=WORKFLOW.USER_EVENT_PARSER_FUNCTION_NAME,
        default_event_parser_function_name=WORKFLOW.DEFAULT_EVENT_PARSER_FUNCTION_NAME,
        action_name=flow.name(connector.subtype),
        type=integration,
        connector_auth=fields(connector.warehouse.auth),
        hrflow_auth=fields(HrFlowWarehouse.auth),
        pull_parameters=fields(origin_parameters),
        push_parameters=fields(target_parameters),
    )


class InvalidConnectorReadmeContent(Exception):
    pass


@dataclass(frozen=True)
class CONNECTOR_README:
    ACTIONS_SECTIONS_REGEXP = (
        r"# 🔌 Connector Actions.+?\|\s*Action\s*\|\s*Description\s*\|.+?\|\s+?<\/p>"
    )


def connector_readme(
    connector: Connector, current_content: t.Optional[str] = None
) -> str:
    if current_content is None:
        return Templates.get_template("connector_readme.md.j2").render(
            connector_name=connector.name.capitalize(),
            connector_subtype=connector.subtype,
            description=connector.description,
            url=connector.url,
            flows=connector.flows,
        )
    else:
        match = re.search(
            CONNECTOR_README.ACTIONS_SECTIONS_REGEXP, current_content, re.DOTALL
        )
        if match is None:
            raise InvalidConnectorReadmeContent(
                "README.md for connector {} does not respect standard format. No"
                " actions section found".format(connector.name)
            )
        updated_actions_content = Templates.get_template(
            "connector_actions.md.j2"
        ).render(
            flows=connector.flows,
        )
        return "{before}{actions}{after}".format(
            before=current_content[: match.start()],
            actions=updated_actions_content,
            after=current_content[match.end() :],
        )


@dataclass(frozen=True)
class CONNECTOR_ACTION:
    class OtherFields(Struct):
        workflow_id: Annotated[
            str,
            Meta(
                description=(
                    "A stable identifier used for persisting in incremental mode"
                )
            ),
        ]
        logics: Annotated[
            t.Optional[LogicsT],
            Meta(
                description=(
                    "A list of functions called in sequence with each"
                    " item pulled from the origin. Each function might either "
                    "return it's argument or None to discard the item. Any item"
                    " discarded is eventually not pushed to the target"
                ),
                extra_json_schema=dict(),
            ),
        ]
        format: Annotated[
            t.Optional[FormatT],
            Meta(
                description=(
                    "A formatting function to apply on items pulled before the push"
                ),
                extra_json_schema=dict(),
            ),
        ] = None
        callback: Annotated[
            t.Optional[CallbackT],
            Meta(
                description=(
                    "Registers a callback function to be called at the "
                    "of a successful execution"
                ),
                extra_json_schema=dict(),
            ),
        ] = None
        persist: Annotated[
            bool,
            Meta(
                description=(
                    "When False has the effect of running "
                    "in dry mode. Items are pulled but not pushed to the target"
                )
            ),
        ] = True
        incremental: Annotated[
            bool, Meta(description="Controls the incremental reading execution mode")
        ] = False

    OTHER_FIELDS = OtherFields


def connector_action(connector: Connector, flow: Flow) -> str:
    import_name = get_import_name(connector)
    action_name = flow.name(connector.subtype)
    if flow.direction is Direction.inbound:
        origin_name = connector.name
        target_name = "HrFlow"
    else:
        origin_name = "HrFlow"
        target_name = connector.name

    connector_aisle = connector.warehouse.get_aisle(flow.entity)
    hrflow_aisle = HrFlowWarehouse.get_aisle(flow.entity)

    # This is only called with a properly validated
    # connector for which below must be true
    assert connector_aisle is not None
    assert hrflow_aisle is not None

    if flow.direction is Direction.inbound:
        origin_aisle = connector_aisle
        target_aisle = hrflow_aisle
    else:
        origin_aisle = hrflow_aisle
        target_aisle = connector_aisle

    # This is only called with a properly validated
    # connector for which below must be true
    assert origin_aisle.read is not None
    assert target_aisle.write is not None

    pull_parameters = origin_aisle.parameters("read", flow.mode)
    push_parameters = target_aisle.parameters("write", flow.mode)

    assert pull_parameters is not None
    assert push_parameters is not None

    return (
        Templates.get_template("action_readme.md.j2")
        .render(
            action_name=action_name,
            origin_name=origin_name,
            target_name=target_name,
            description=flow.description(connector.name),
            origin_endpoint=origin_aisle.read.endpoints.for_mode(flow.mode),
            target_endpoint=target_aisle.write.endpoints.for_mode(flow.mode),
            connector_name=connector.name,
            connector_auth_fields=template_fields(connector.warehouse.auth),
            hrflow_auth_fields=template_fields(HrFlowWarehouse.auth),
            pull_fields=template_fields(pull_parameters),
            push_fields=template_fields(push_parameters),
            other_fields=template_fields(CONNECTOR_ACTION.OTHER_FIELDS),
            main_module=MAIN_IMPORT_NAME.get(),
            import_name=import_name,
        )
        .replace(",\n    \n)", "\n)")
    )


@dataclass(frozen=True)
class CONNECTOR_STUB:
    pass


def connector_stub(
    connector: Connector,
) -> str:
    return Templates.get_template("connector.pyi.j2").render(
        import_name=get_import_name(connector),
        actions=[flow.name(connector.subtype) for flow in connector.flows],
    )



import enum
import json
import logging
import os
import re
import subprocess
import typing as t
from contextvars import ContextVar
from datetime import datetime
from pathlib import Path

from jinja2 import Template
from pydantic import BaseModel
from pydantic.fields import ModelField
from typing_extensions import TypeGuard

from hrflow_connectors.v1.core.common import ALL_TARGET_CONNECTORS_LIST_PATH
from hrflow_connectors.v1.core.connector import (
    MAIN_IMPORT_NAME,
    Connector,
    get_import_name,
)
from hrflow_connectors.v1.core.templates import Templates
from hrflow_connectors.v2 import __CONNECTORS__ as __CONNECTORS__V2
from hrflow_connectors.v2.core.connector import Connector as V2Connector

logger = logging.getLogger(__name__)
CONNECTORS_DIRECTORY = Path(__file__).parent.parent / "connectors"

ACTIONS_SECTIONS_REGEXP = (
    r"# 🔌 Connector Actions.+?\|\s*Action\s*\|\s*Description\s*\|.+?\|\s+?<\/p>"
)

GIT_UPDATE_EXCLUDE_PATTERN = r"(notebooks/\.gitkeep|mappings/format/\.gitkeep|README\.md|test\-config\.yaml|logo\.png|docs/)"
GIT_UPDATE_TIMEOUT = 5
# The git log command was updated with the
# of following arguments to discard the commits
# which are related to v2 migration
# --grep "[v1-v2-migration]" --invert-grep
# In case no commit pass the condition below then
# then "pre_v2_updated_at" from connectors.json is used
GIT_UPDATE_DATE = """
git ls-tree -r --name-only HEAD {base_connector_path}/{connector} | while read filename; do
  echo "$(git log -1 --grep "[v1-v2-migration]" --invert-grep --format="%cI" -- $filename) $filename"
done
"""

HRFLOW_CONNECTORS_REMOTE_URL = "https://github.com/Riminder/hrflow-connectors"
USE_REMOTE_REV: ContextVar[t.Optional[str]] = ContextVar("USE_REMOTE_REV", default=None)
BASE_CONNECTOR_PATH: ContextVar[t.Optional[str]] = ContextVar(
    "BASE_CONNECTOR_PATH", default="src/hrflow_connectors/v1/connectors/"
)
PREMIUM_STATUS = ":lock: Premium"
PREMIUM_README_LINK = "https://forms.gle/pokoE9pAjSVSFtCe7"
OPENSOURCE_STATUS = ":book: Open source"

V2_CONNECTORS_BY_NAME = {connector.name: connector for connector in __CONNECTORS__V2}


class InvalidConnectorReadmeFormat(Exception):
    pass


class TemplateField(BaseModel):
    name: str
    type: str
    required: bool
    description: str
    example: str
    default: str


def field_example(field: ModelField) -> str:
    if callable(field.default):
        return "lambda *args, **kwargs: None # Put your code logic here"

    if field.default is not None:
        if isinstance(field.default, str):
            return '"{}"'.format(field.default)
        return str(field.default)

    if field.default_factory is not None:
        return str(field.default_factory())

    field_type = field.outer_type_
    if isinstance(field_type, enum.EnumMeta):
        return '"{}"'.format(list(field_type)[0].value)

    if field_type is str:
        return '"your_{}"'.format(field.name)

    if field_type in [int, float, bool]:
        return str(field_type())

    return "***"


def field_default(field: ModelField, documentation_path: Path) -> str:
    if callable(field.default):
        filepath = os.path.relpath(
            field.default.__code__.co_filename, documentation_path
        )
        if (
            "site-packages/hrflow_connectors/" in filepath
            and USE_REMOTE_REV.get() is not None
        ):
            filepath = "{}/tree/{}/src/hrflow_connectors/{}".format(
                HRFLOW_CONNECTORS_REMOTE_URL,
                USE_REMOTE_REV.get(),
                filepath.split("/hrflow_connectors/")[-1],
            )
        return "[`{}`]({}#L{})".format(
            field.default.__code__.co_name,
            filepath,
            field.default.__code__.co_firstlineno,
        )

    if field.default_factory is not None:
        return str(field.default_factory())

    return str(field.default)


def field_type(field: ModelField) -> str:
    if field.outer_type_ in [int, float, str, bool]:
        return field.outer_type_.__name__
    if isinstance(field.outer_type_, enum.EnumMeta):
        return "str"
    return str(field.outer_type_)


def get_template_fields(
    fields: t.List[ModelField], documentation_path: Path
) -> t.List[TemplateField]:
    return [
        TemplateField(
            name=field.name,
            type=field_type(field),
            required=field.required,
            description=field.field_info.description or "",
            example=field_example(field),
            default=field_default(field, documentation_path),
        )
        for field in fields
        if not field.field_info.const
        and field.field_info.extra.get("skip_from_docs", False) is False
    ]


def py_37_38_compat_patch(content: str) -> str:
    """
    The way t.Optional[T] is stringified is different accross supported python versions:
        - Python 3.7, 3.8 --> typing.Union[T, NoneType]
        - Python >= 3.9 --> t.Optional[T]
    This creates inconsistency when generating the doc with accross these versions.
    This function changes any older string versions to match with >=3.9
    """
    return re.sub(
        r"Union\[([\w\.]+), NoneType\]",
        lambda match: f"Optional[{match.group(1)}]",
        content,
    )


def ensure_gitkeep(directory: Path, gitkeep_filename: str = ".gitkeep") -> None:
    gitkeep_file = directory / gitkeep_filename
    create_empty_file = True

    if directory.is_dir():
        for child in directory.iterdir():
            if not child.name == gitkeep_file.name:
                create_empty_file = False
                try:
                    gitkeep_file.unlink()
                except FileNotFoundError:
                    pass
                break
    else:
        directory.mkdir(parents=True)

    if create_empty_file:
        gitkeep_file.touch()


def connector_is_v2(
    connector: t.Union[Connector, V2Connector],
) -> TypeGuard[V2Connector]:
    return isinstance(connector, V2Connector)


def update_root_readme(
    connectors: t.List[Connector],
    target_connectors: t.List[t.Dict],
    root: Path,
    root_template: Template,
) -> t.Dict:
    connector_by_name = {connector.model.name: connector for connector in connectors}
    all_connectors = sorted(
        [
            {
                **connector,
                "object": V2_CONNECTORS_BY_NAME.get(
                    connector["name"], connector_by_name.get(connector["name"])
                ),
            }
            for connector in target_connectors
        ],
        key=lambda c: c["name"].lower(),
    )

    line_pattern = (
        "| [**{name}**]({readme_link}) | {type} | {status} |"
        " {release_date} | {updated_at} |"
    )
    opensource_connectors_table = ""
    opensource_jobboards_table = ""
    premium_connectors_table = ""
    premium_jobboards_table = ""
    for connector in all_connectors:
        if connector["object"] is None:
            updated_listing = line_pattern.format(
                name=connector["name"],
                readme_link=PREMIUM_README_LINK,
                type=connector["type"],
                status=PREMIUM_STATUS,
                release_date="",
                updated_at="",
            )
            if connector["type"] == "Job Board":
                premium_jobboards_table += updated_listing + "\n"
            else:
                premium_connectors_table += updated_listing + "\n"
        else:
            connector_object = t.cast(
                t.Union[Connector, V2Connector], connector["object"]
            )
            if connector_is_v2(connector_object):
                name = connector_object.name
                subtype = connector_object.subtype
                connector_type = connector_object.type
                base_connector_path = (
                    BASE_CONNECTOR_PATH.get().rstrip("/").replace("v1", "v2")
                )
            else:
                name = connector_object.model.name
                subtype = connector_object.model.subtype
                connector_type = connector_object.model.type
                base_connector_path = BASE_CONNECTOR_PATH.get().rstrip("/")

            result = subprocess.run(
                GIT_UPDATE_DATE.format(
                    connector=subtype,
                    base_connector_path=base_connector_path,
                ),
                shell=True,
                text=True,
                capture_output=True,
                timeout=GIT_UPDATE_TIMEOUT,
            )
            if result.stderr:
                raise Exception(
                    "Subprocess run for Git update dates failed for connector {} with"
                    " errors {}".format(subtype, result.stderr)
                )
            filtered = [
                line.split(" ")[0]
                for line in filter(
                    lambda line: not re.search(GIT_UPDATE_EXCLUDE_PATTERN, line),
                    result.stdout.strip().splitlines(),
                )
            ]
            non_empty = [entry for entry in filtered if entry != ""]
            if len(non_empty) == 0:
                updated_at = connector["pre_v2_updated_at"]
            else:
                updated_at = datetime.fromisoformat(
                    max(
                        non_empty,
                        key=lambda d: datetime.fromisoformat(d.replace("Z", "+00:00")),
                    ).replace("Z", "+00:00")
                ).strftime("%d/%m/%Y")

            updated_listing = line_pattern.format(
                name=name,
                readme_link="./{base_connector_path}/{connector}/README.md".format(
                    base_connector_path=base_connector_path,
                    connector=subtype,
                ),
                type=connector_type.value,
                status=OPENSOURCE_STATUS,
                release_date=f'*{connector["release_date"]}*',
                updated_at=f"*{updated_at}*",
            )

            if connector["type"] == "Job Board":
                opensource_jobboards_table += updated_listing + "\n"
            else:
                opensource_connectors_table += updated_listing + "\n"

    readme = root / "README.md"
    readme_content = root_template.render(
        opensource_connectors_table=opensource_connectors_table.strip("\n"),
        opensource_jobboards_table=opensource_jobboards_table.strip("\n"),
        premium_connectors_table=premium_connectors_table.strip("\n"),
        premium_jobboards_table=premium_jobboards_table.strip("\n"),
    )
    readme_content = py_37_38_compat_patch(readme_content)
    readme.write_bytes(readme_content.encode())


KEEP_EMPTY_FOLDER = ".gitkeep"


def generate_docs(
    connectors: t.List[Connector],
    target_connectors: t.Optional[t.List[t.Dict]] = None,
    connectors_directory: Path = CONNECTORS_DIRECTORY,
    root_template: Template = Templates.get_template("root_readme.md.j2"),
) -> None:
    if target_connectors is None:
        with open(ALL_TARGET_CONNECTORS_LIST_PATH, "r") as f:
            target_connectors = json.load(f)
    update_root_readme(
        connectors=connectors,
        target_connectors=target_connectors,
        root=connectors_directory.parent.parent.parent.parent,
        root_template=root_template,
    )
    for connector in connectors:
        model = connector.model
        connector_directory = connectors_directory / model.subtype
        if not connector_directory.is_dir():
            logging.error(
                "Skipping documentation for {}: no directory found at {}".format(
                    model.name, connector_directory
                )
            )
            continue

        import_name = get_import_name(connector)

        readme = connector_directory / "README.md"
        if readme.exists() is False:
            readme_content = Templates.get_template("connector_readme.md.j2").render(
                connector_name=model.name.replace(" ", "").capitalize(),
                description=model.description,
                url=model.url,
                actions=model.actions,
            )
            readme_content = py_37_38_compat_patch(readme_content)
            readme.write_bytes(readme_content.encode())
        else:
            readme_content = readme.read_text()
            match = re.search(ACTIONS_SECTIONS_REGEXP, readme_content, re.DOTALL)
            if match is None:
                raise InvalidConnectorReadmeFormat(
                    "README.md for connector {} does not respect standard format. No"
                    " actions section found".format(model.name)
                )
            updated_actions_content = Templates.get_template(
                "connector_actions.md.j2"
            ).render(
                actions=model.actions,
            )
            updated_readme_content = "{before}{actions}{after}".format(
                before=readme_content[: match.start()],
                actions=updated_actions_content,
                after=readme_content[match.end() :],
            )
            updated_readme_content = py_37_38_compat_patch(updated_readme_content)
            readme.write_bytes(updated_readme_content.encode())

        notebooks_directory = connector_directory / "notebooks"
        ensure_gitkeep(notebooks_directory, KEEP_EMPTY_FOLDER)

        format_mappings_directory = connector_directory / "mappings" / "format"
        ensure_gitkeep(format_mappings_directory, KEEP_EMPTY_FOLDER)

        if len(model.actions) > 0:
            action_docs_directory = connector_directory / "docs"
            if not action_docs_directory.is_dir():
                action_docs_directory.mkdir()
            for action in model.actions:
                action_name = action.name.value
                action_fields = get_template_fields(
                    fields=action.parameters.__fields__.values(),
                    documentation_path=action_docs_directory,
                )
                origin_fields = get_template_fields(
                    fields=action.origin.read.parameters.__fields__.values(),
                    documentation_path=action_docs_directory,
                )
                target_fields = get_template_fields(
                    fields=action.target.write.parameters.__fields__.values(),
                    documentation_path=action_docs_directory,
                )
                action_documentation_content = Templates.get_template(
                    "action_readme.md.j2"
                ).render(
                    main_module=MAIN_IMPORT_NAME.get(),
                    import_name=import_name,
                    action_name=action_name,
                    description=action.description,
                    action_fields=action_fields,
                    origin_name=action.origin.name,
                    origin_fields=origin_fields,
                    origin_endpoints=action.origin.read.endpoints,
                    target_name=action.target.name,
                    target_fields=target_fields,
                    target_endpoints=action.target.write.endpoints,
                )
                action_documentation_content = py_37_38_compat_patch(
                    action_documentation_content
                )
                action_documentation = action_docs_directory / "{}.md".format(
                    action_name
                )
                action_documentation.write_bytes(action_documentation_content.encode())
